import xml.etree.ElementTree as ET
from ase.atoms import Atoms
import numpy as np
from ase.units import Bohr, eV, J
from .constants import gyromagnetic_ratio
from ase.data import atomic_masses


class BaseSpinModelParser(object):
    """
    SpinModelParser: a general model for spin model file parser.
    """

    def __init__(self, fname):
        self._fname = fname
        self.atoms = None
        self.damping_factors = []
        self.gyro_ratios = []
        self.index_spin = []
        self.cell = None
        self.zions = []
        self.masses = []
        self.positions = []
        self.spinat = []
        self._exchange = {}
        self._dmi = {}
        self._bilinear={}
        self._parse(fname)
        self.lattice = Atoms(
            positions=self.positions, masses=self.masses, cell=self.cell)

    def _parse(self, fname):
        raise NotImplementedError("parse function not implemented yet")

    def get_atoms(self):
        return self.atoms

    def _spin_property(self, prop):
        return [
            prop[i] for i in range(len(self.index_spin))
            if self.index_spin[i] > 0
        ]

    @property
    def spin_positions(self):
        return np.array(self._spin_property(self.positions), dtype='float')

    @property
    def spin_zions(self):
        return np.array(self._spin_property(self.zions), dtype='int')

    @property
    def spin_spinat(self):
        return np.array(self._spin_property(self.spinat), dtype='float')

    @property
    def spin_damping_factors(self):
        return np.array(
            self._spin_property(self.damping_factors), dtype='float')

    @property
    def spin_gyro_ratios(self):
        return np.array(self._spin_property(self.gyro_ratios), dtype='float')

    def get_index_spin(self):
        return self.index_spin

    def exchange(self, isotropic=True):
        if isotropic:
            iso_jdict = {}
            for key, val in self._exchange.items():
                iso_jdict[key] = val[0]
            return iso_jdict
        else:
            return self._exchange

    @property
    def dmi(self):
        return self._dmi

    @property
    def bilinear(self):
        return self._bilinear


    @property
    def has_exchange(self):
        return bool(len(self._exchange))

    @property
    def has_dmi(self):
        return bool(len(self._dmi))

    @property
    def has_bilinear(self):
        return bool(len(self._bilinear))



class SpinXmlWriter(object):
    def _write(self, model, fname):
        root = ET.Element("System_definition")
        unitcell = model.cell.reshape((3, 3)) / Bohr
        uctext = "\t\n".join(
            ["\t".join(["%.5e" % x for x in ui]) for ui in unitcell])
        uc = ET.SubElement(root, "unitcell", units="bohrradius")
        uc.text = uctext
        nspin = len(model.zion)
        model._map_to_magnetic_only()
        id_spin = [-1] * nspin
        counter = 0
        for i in np.array(model.magsites, dtype=int):
            counter += 1
            id_spin[int(i)] = counter
        for i, z in enumerate(model.zion):
            atom = ET.SubElement(
                root,
                "atom",
                damping_factor="%.5e" % (1.0),
                gyroratio="%.5e" % (model.gyro_ratio[i] / gyromagnetic_ratio),
                mass="%.5e" % atomic_masses[z],
                index_spin="%d" % id_spin[i],
                massunits="atomicmassunit")
            pos = ET.SubElement(atom, "position", units="bohrradius")
            pos.text = "%.5e\t%.5e\t%.5e" % tuple(model.xcart[i]/Bohr)
            spinat = ET.SubElement(atom, "spinat")
            spinat.text = "%.5e\t%.5e\t%.5e" % tuple(model.spinat[i])

        if model.has_exchange:
            exc = ET.SubElement(root, "spin_exchange_list", units="eV")
            ET.SubElement(exc,
                          "nterms").text = "%s" % (len(model.exchange_Jdict))
            for key, val in model.exchange_Jdict.items():
                exc_term = ET.SubElement(exc, "spin_exchange_term", units="eV")
                ET.SubElement(exc_term, "ijR").text = "%d %d %d %d %d" % (
                    key[0] + 1, key[1] + 1, key[2][0], key[2][1], key[2][2])
                #ET.SubElement(exc_term,
                #              "data").text = "%.5e \t %.5e \t %.5e" % (
                #                  val * J / eV, val * J / eV, val * J / eV)

                try:  # if val is a iterable
                    ET.SubElement(
                        exc_term, "data").text = "%.5e \t %.5e \t %.5e" % (
                            val[0] * J / eV, val[1] * J / eV, val[2] * J / eV)
                except Exception:
                    ET.SubElement(exc_term,
                                  "data").text = "%.5e \t %.5e \t %.5e" % (
                                      val * J / eV, val * J / eV, val * J / eV)
        if model.has_dmi:
            dmi = ET.SubElement(root, "spin_DMI_list", units="eV")
            ET.SubElement(dmi, "nterms").text = "%d" % len(model.dmi_ddict)
            for key, val in model.dmi_ddict.items():
                dmi_term = ET.SubElement(dmi, "spin_DMI_term")
                ET.SubElement(dmi_term, "ijR").text = "%d %d %d %d %d" % (
                    key[0] + 1, key[1] + 1, key[2][0], key[2][1], key[2][2])
                ET.SubElement(
                    dmi_term, "data").text = "%.5e \t %.5e \t %.5e" % (
                        val[0] * J / eV, val[1] * J / eV, val[2] * J / eV)

        if model.has_uniaxial_anistropy:
            uni = ET.SubElement(root, "spin_uniaxial_SIA_list", units="eV")
            ET.SubElement(uni, "nterms").text = "%d" % len(model.k1)
            for i, k1 in enumerate(model.k1):
                uni_term = ET.SubElement(uni, "spin_uniaxial_SIA_term")
                ET.SubElement(uni_term, "i").text = "%d " % (i + 1)
                ET.SubElement(uni_term,
                              "amplitude").text = "%.5e" % (k1 * J / eV)
                ET.SubElement(
                    uni_term,
                    "direction").text = "%.5e \t %.5e \t %.5e " % tuple(
                        model.k1dir[i])

        if model.has_bilinear:
            bi = ET.SubElement()
            bilinear = ET.SubElement(root, "spin_bilinear_list", units="eV")
            ET.SubElement(bilinear,
                          "nterms").text = "%d" % len(model.bilinear_J_dict)
            for key, val in model.bilinear_Jdict.items():
                bilinear_term = ET.SubElement(bilinear, "spin_bilinear_term")
                ET.SubElement(bilinear_term, "ijR").text = "%d %d %d %d %d" % (
                    key[0] + 1, key[1] + 1, key[2][0], key[2][1], key[2][2])
                ET.SubElement(bilinear_term, "data").text = '\t'.join(
                    ["%.5e" % (x * J / eV) for x in val])

        tree = ET.ElementTree(root)
        tree.write(fname)


class SpinXmlParser(BaseSpinModelParser):
    """
    parser to spin xml file.
    """

    def _parse(self, fname):
        tree = ET.parse(fname)
        root = tree.getroot()

        self.atoms = Atoms()
        uc = root.find('unit_cell')
        cell = np.array([float(x) for x in uc.text.strip().split()])
        self.cell = np.reshape(cell, (3, 3)) * Bohr

        for a in root.findall('atom'):
            if 'damping_factor' in a.attrib:
                self.damping_factors.append(float(a.attrib['damping_factor']))
            else:
                print("warning: damping factor not found")
                self.damping_factors.append(0.0)

            if 'gyroratio' in a.attrib:
                self.gyro_ratios.append(float(a.attrib['gyroratio']))
            else:
                print("warning: gyroratio not found")
                self.gyro_ratios.append(0.0)

            if 'index_spin' in a.attrib:
                self.index_spin.append(int(a.attrib['index_spin']))
            else:
                print("warning: index_spin not found")
                self.index_spin.append(-1)

            if 'mass' in a.attrib:
                self.masses.append(float(a.attrib['mass']))
            else:
                print("warning: mass not found")
                self.masses.append(1)

            if 'zion' in a.attrib:
                self.zions.append(float(a.attrib['zion']))
            else:
                #print("warning: zion not found")
                self.zions.append(1)

            p = a.find('position')
            if p is not None:
                position = np.array(tuple(map(float,
                                              p.text.strip().split()))) * Bohr
                self.positions.append(position)
            else:
                raise Exception("position not found for one atom")

            p = a.find('spinat')
            if p is not None:
                spin = np.array(tuple(map(float, p.text.strip().split())))
                self.spinat.append(spin)
            else:
                raise Exception("spinat not found for one atom")

        exch = root.find('spin_exchange_list')
        n_exch = int(exch.find('nterms').text)

        for exc in exch.findall('spin_exchange_term'):
            ijR = [int(x) for x in exc.find('ijR').text.strip().split()]
            i, j, R0, R1, R2 = ijR
            val = [float(x) for x in exc.find('data').text.strip().split()]
            self._exchange[(i - 1, j - 1, (R0, R1,
                                           R2))] = np.array(val) * eV / J
        assert len(
            self._exchange
        ) == n_exch, "Number of exchange terms different from nterms in xml file"

        dmi = root.find('spin_DMI_list')
        if dmi is not None:
            n_dmi= int(dmi.find('nterms').text)
            for dm in dmi.findall('spin_DMI_term'):
                ijR = [int(x) for x in dm.find('ijR').text.strip().split()]
                i, j, R0, R1, R2 = ijR
                val = [float(x) for x in dm.find('data').text.strip().split()]
                self._dmi[(i - 1, j - 1, (R0, R1,
                                               R2))] = np.array(val) * eV / J
            assert len(
                self._dmi
            ) == n_dmi, f"Number of dmi terms {len(self._dmi)} different from nterms in xml file {n_dmi}"
     
        bil = root.find('spin_bilinear_list')
        if bil is not None:
            n_bil= int(bil.find('nterms').text)
            for bi in bil.findall('spin_bilinear_term'):
                ijR = [int(x) for x in bi.find('ijR').text.strip().split()]
                i, j, R0, R1, R2 = ijR
                val = [float(x) for x in bi.find('data').text.strip().split()]
                self._bilinear[(i - 1, j - 1, (R0, R1,
                                               R2))] = np.array(val).reshape((3,3)) * eV / J
            assert len(
                self._bilinear
            ) == n_bil, f"Number of bilinear terms {len(self._bil)} different from nterms in xml file {n_bil}"
     
        



import numpy as np
import xml.etree.cElementTree as ET
from xml.dom import minidom
from ase.data import atomic_masses
from ase.units import eV, Hartree, Bohr, Ry, J
import os
from collections import Iterable, namedtuple
from itertools import groupby
from TB2J.utils import symbol_number
import pickle



def write_multibinit_inp(cls, path='TB2J_results/Multibinit'):
    if not os.path.exists(path):
        os.makedirs(path)
    # files file
    filesfname = os.path.join(path, 'mb.files')
    with open(filesfname, 'w') as myfile:
        myfile.write("""mb.in
mb.out
exchange.xml
""")
    inputfname = os.path.join(path, 'mb.in')
    # template for input file
    with open(inputfname, 'w') as myfile:
        infile_template = """
prt_model = 0

#--------------------------------------------------------------
#Monte carlo / molecular dynamics
#--------------------------------------------------------------
dynamics =  0    ! disable molecular dynamics

ncell =   16 16 16  ! size of supercell. 
#-------------------------------------------------------------
#Spin dynamics
#------------------------------------------------------------
spin_dynamics=1  ! enable spin dynamics
spin_mag_field= 0.0 0.0 0.0   ! external magnetic field
spin_ntime_pre =10000          ! warming up steps. 
spin_ntime =10000             ! number of steps. 
spin_nctime=100               ! number of time steps between two nc file write
spin_dt=5e-16 s               ! time step. 
spin_init_state = 1           ! random initial spin

spin_temperature=0.0

spin_var_temperature=1        ! switch on variable temperature calculation
spin_temperature_start=0      ! starting point of temperature
spin_temperature_end=500      ! ending point of temperature. 
spin_temperature_nstep=6      ! number of temperature steps.

spin_sia_add = 1              ! add a single ion anistropy (SIA) term?
spin_sia_k1amp = 1e-6         ! amplitude of SIA (in Ha), how large should be used?
spin_sia_k1dir = 0.0 0.0 1.0  ! direction of SIA

spin_calc_thermo_obs = 1      ! calculate thermodynamics related observables
        """
        myfile.write(infile_template)

def write_xml(cls, fname):
    root = ET.Element("System_definition")
    try:
        unitcell = cls.atoms.get_cell().reshape((3, 3)) / Bohr
    except Exception:  # ase>3.18 cell is not a array.
        unitcell = cls.atoms.get_cell().array.reshape((3, 3)) / Bohr
    uctext = "\t\n".join(
        ["\t".join(["%.5e" % x for x in ui]) for ui in unitcell])
    uc = ET.SubElement(root, "unit_cell", units="bohrradius")
    uc.text = uctext
    #ET.SubElement(
    #    root, "unit_cell", units="bohrradius").text = "\t".join(
    #        list(
    #            "%.5e" % x
    #            for x in np.array(cls.atoms.get_cell()).flatten() / Bohr))
    natom = len(cls.atoms)
    #cls._map_to_magnetic_only()
    #id_spin = [-1] * natom
    #counter = 0
    #for i in np.array(cls.magsites, dtype=int):
    #    counter += 1
    #    id_spin[int(i)] = counter

    for i in range(natom):
        atom = ET.SubElement(
            root,
            "atom",
            mass="%.5e" % cls.atoms.get_masses()[i],
            massunits="atomicmassunit",
            damping_factor="%.5f" % cls.
            damping[i],  # TODO remove this. damping factor is not local.
            gyroratio="%.5e" % cls.gyro_ratio[i],
            index_spin="%d" % (cls.index_spin[i] + 1),  # +1 in fortran
        )
        pos = ET.SubElement(atom, "position", units="bohrradius")
        pos.text = "%.5e\t%.5e\t%.5e" % tuple(
            cls.atoms.get_positions()[i] / Bohr)
        spinat = ET.SubElement(atom, "spinat")
        spinat.text = "%.5e\t%.5e\t%.5e" % tuple(cls.spinat[i])

    if cls.has_exchange:
        exc = ET.SubElement(root, "spin_exchange_list", units="eV")
        ET.SubElement(exc,
                        "nterms").text = "%s" % (len(cls.exchange_Jdict))
        for key, val in cls.exchange_Jdict.items():
            R, i, j = key
            exc_term = ET.SubElement(exc, "spin_exchange_term")
            ET.SubElement(
                exc_term,
                "ijR").text = "%d %d %d %d %d" % (i + 1, j + 1, R[0], R[1],
                                                    R[2])
            try:
                ET.SubElement(exc_term,
                                "data").text = "%.5e \t %.5e \t %.5e" % (
                                    val[0], val[1], val[2])
            except Exception:
                ET.SubElement(
                    exc_term,
                    "data").text = "%.5e \t %.5e \t %.5e" % (val, val, val)
    if cls.has_dmi:
        dmi = ET.SubElement(root, "spin_DMI_list", units="eV")
        ET.SubElement(dmi, "nterms").text = "%d" % len(cls.dmi_ddict)
        for key, val in cls.dmi_ddict.items():
            R, i, j = key
            dmi_term = ET.SubElement(dmi, "spin_DMI_term")
            ET.SubElement(
                dmi_term,
                "ijR").text = "%d %d %d %d %d" % (i + 1, j + 1, R[0], R[1],
                                                    R[2])
            ET.SubElement(
                dmi_term,
                "data").text = "%.5e \t %.5e \t %.5e" % (val[0], val[1],
                                                            val[2])

    if cls.has_uniaxial_anistropy:
        uni = ET.SubElement(root, "spin_uniaxial_SIA_list", units="eV")
        ET.SubElement(uni, "nterms").text = "%d" % len(cls.k1)
        for i, k1 in enumerate(cls.k1):
            uni_term = ET.SubElement(uni, "spin_uniaxial_SIA_term")
            ET.SubElement(uni_term, "i").text = "%d " % (i + 1)
            ET.SubElement(uni_term, "amplitude").text = "%.5e" % k1
            ET.SubElement(
                uni_term,
                "direction").text = "%.5e \t %.5e \t %.5e " % tuple(
                    cls.k1dir[i])

    if cls.has_bilinear:
        bilinear = ET.SubElement(root, "spin_bilinear_list", units="eV")
        ET.SubElement(bilinear,
                        "nterms").text = "%d" % len(cls.Jani_dict)
        for key, val in cls.Jani_dict.items():
            bilinear_term = ET.SubElement(bilinear, "spin_bilinear_term")
            ET.SubElement(bilinear_term, "ijR").text = "%d %d %d %d %d" % (
                key[1] + 1, key[2] + 1, key[0][0], key[0][1], key[0][2])
            ET.SubElement(bilinear_term, "data").text = '\t'.join(
                ["%.5e" % x for x in val.flatten()])
    if cls.description is not None:
        description = ET.SubElement(root, "description")
        description.text = cls.description

    #tree = ET.ElementTree(root)
    #tree.write(fname)
    with open(fname, 'w') as myfile:
        myfile.write(
            minidom.parseString(
                ET.tostring(root)).toprettyxml(indent="\t"))


def write_multibinit(cls, path):
    write_multibinit_inp(cls, path=path)
    xmlfname = os.path.join(path, 'exchange.xml')
    write_xml(cls, xmlfname)




import numpy as np
import xml.etree.cElementTree as ET
from xml.dom import minidom
from ase.data import atomic_masses
from ase.units import eV, Hartree, Bohr, Ry, J
import os
from collections import Iterable, namedtuple
from itertools import groupby
from TB2J.utils import symbol_number
import pickle


def write_uppasd(cls, path='TB2J_results/UppASD'):
    if not os.path.exists(path):
        os.makedirs(path)
    cls.write_uppasd_posfile(os.path.join(path, 'posfile'))
    cls.write_uppasd_momfile(os.path.join(path, 'momfile'))
    cls.write_uppasd_exchange(os.path.join(path, 'jASD1'))
    cls.write_uppasd_infile(os.path.join(path, 'input'))

def write_uppasd_posfile(cls, fname):
    with open(fname, 'w') as myfile:
        natom = len(cls.atoms)
        for i in range(natom):
            text = ""
            id_spin = cls.index_spin[i]
            if id_spin > -1:
                pos = cls.atoms.get_scaled_positions()[i]
                text = "{id_atom} {id_spin} {pos_x} {pos_y} {pos_z}\n".format(
                    id_atom=id_spin + 1,
                    id_spin=id_spin + 1,
                    pos_x=pos[0],
                    pos_y=pos[1],
                    pos_z=pos[2],
                )
                myfile.write(text)

def write_uppasd_momfile(cls, fname):
    with open(fname, 'w') as myfile:
        natom = len(cls.atoms)
        for i in range(natom):
            text = ""
            id_spin = cls.index_spin[i]
            if id_spin > -1:
                pos = cls.atoms.get_scaled_positions()[i]
                ms = np.sqrt(np.sum(np.array(cls.spinat[i])**2))
                spin = np.array(cls.spinat[i]) / ms
                text = "{id_atom} {id_spin} {ms} 0.0 0.0 1.0\n".format(
                    id_atom=id_spin + 1, id_spin=id_spin + 1, ms=ms)
                myfile.write(text)

def write_uppasd_exchange(cls, fname):
    with open(fname, 'w') as myfile:
        nexch = len(cls.exchange_Jdict.items())
        myfile.write("{num_interactions} {type_exchange}\n".format(
            num_interactions=nexch, type_exchange=0))

        counter = -1
        for key, val in cls.exchange_Jdict.items():
            counter += 1  # starts at 0
            R, i, j = key
            pos = cls.atoms.get_positions()
            d = np.dot(np.array(R),
                       cls.atoms.get_cell()) + pos[j] - pos[i]
            myfile.write("{i} {j} {Rx} {Ry} {Rz} {Jij}\n".format(
                IID=counter,
                i=i + 1,
                j=j + 1,
                Rx=d[0],
                Ry=d[1],
                Rz=d[2],
                Jij=val * 1e3 / Ry))  # mRy

def write_uppasd_infile(cls, fname):
    tmpl = """ simid Unamed
    ncell 12 12 12
    BC P P P
    cell
    Sym 0
    posfile ./posfile
    momfile ./momfile
    exchange ./jASD1
    #anistropy ./kfile

    do_ralloy 0
    Mensemble 1
    tseed 4499
    maptype 1
    SDEalgh 1
    Initmag 1
    ip_mode M
    ip_mcanneal 1
    10000 300 1.00e-16 0.3

    mode M
    temp 300
    mcNstep 50000
    Nstep 50000
    damping 0.1
    timestep 1.0e-16

    do_avrg Y

    do_cumu Y
    cumu_step 50
    cumu_buff 10

    do_tottraj N
    tottraj_step 1000

    plotenergy 1

    do_sc C
    do_ams Y
    do_magdos Y
    magdos_freq 200
    magdos_sigma 30
    qpoints C

    do_stiffness Y
    eta_max 12
    eta_min 6
    alat 2.83e-10
    """
    with open(fname, 'w') as myfile:
        myfile.write(tmpl)




from collections import defaultdict
import os
import shutil
import numpy as np
import time
from TB2J.green import TBGreen
from TB2J.utils import symbol_number, read_basis, kmesh_to_R
from TB2J.myTB import MyTB
from ase.io import read
from TB2J.utils import auto_assign_basis_name
from TB2J.io_exchange import SpinIO
from functools import lru_cache
from .exchange import ExchangeCL
from .utils import simpson_nonuniform, trapezoidal_nonuniform
from pathos.multiprocessing import ProcessPool
from ase.dft.kpoints import monkhorst_pack
from TB2J.exchangeCL2 import ExchangeCL2


def find_index_k(kpts, q):
    """
    for one q point, find the indices of k'=k+q inside the space of k.
    kpts: A list of kpoints, should be a regular grid e.g. Monkhorst-pack
    q: one q point, which should be one point in the regular grid.
    """
    kpts_p = np.mod(kpts + 1e-9, 1)
    jkpts = np.zeros(len(kpts), dtype=int)
    for ik, k in enumerate(kpts):
        jkpts[ik] = np.argmin(
            np.linalg.norm(np.mod(k + q + 1e-9, 1)[None, :] - kpts_p, axis=1))
    if len(jkpts) != len(set(jkpts)):
        raise ValueError(
            "Cannot find all the k+q point. Please check if the k-mesh and the q-point is compatible."
        )
    return jkpts


class ExchangeCLQspace(ExchangeCL2):
    def _prepare(self):
        self.nmagatom = len(self.ind_mag_atoms)
        #self.qmesh=[3,3,3]
        self.qmesh = self.kmesh
        self.qpts = monkhorst_pack(size=self.qmesh)
        self.Rqlist = kmesh_to_R(self.qmesh)

        self.nqpt = len(self.qpts)
        self.ncontour = len(self.contour.path)
        self.Jqe_list = np.zeros(
            (self.ncontour, self.nqpt, self.nmagatom, self.nmagatom),
            dtype=complex)
        self.Kqe_list = np.zeros_like(self.Jqe_list)
        self.Xqe_list = np.zeros_like(self.Jqe_list)
        self.get_rho_atom()

    def get_rho_atom(self):
        """
        charges and spins from density matrices
        """
        self.charges = np.zeros(len(self.atoms), dtype=float)
        self.spinat = np.zeros((len(self.atoms), 3), dtype=float)

        rho_up = self.Gup.get_density()
        rho_dn = self.Gdn.get_density()
        for iatom in self.orb_dict:
            iorb = self.iorb(iatom)
            tup = np.sum(rho_up[iorb])
            tdn = np.sum(rho_dn[iorb])
            self.charges[iatom] = tup + tdn
            self.spinat[iatom, 2] = tup - tdn

    def Gk_atom(self, Gk, iatom, jatom):
        return self.GR_atom(Gk, iatom, jatom)

    @property
    def kpts(self):
        return self.Gup.kpts

    @property
    def nkpts(self):
        return len(self.Gup.kpts)

    @lru_cache(maxsize=None)
    def get_ikplusq(self, q):
        return find_index_k(self.kpts, q)

    def get_all_A(self):
        # prepare Gk
        Gk_up = np.zeros((self.nkpts, self.Gup.norb, self.Gup.norb),
                         dtype=complex)
        Gk_dn = np.zeros((self.nkpts, self.Gup.norb, self.Gup.norb),
                         dtype=complex)

        for ie, energy in enumerate(self.contour.path):
            for ik, _ in enumerate(self.kpts):
                Gk_up[ik] = self.Gup.get_Gk(ik, energy)
                Gk_dn[ik] = self.Gdn.get_Gk(ik, energy)

            for iq, q in enumerate(self.qpts):
                ikplusq_list = self.get_ikplusq(tuple(q))

                for ik, ikq in enumerate(ikplusq_list):
                    #Kq= Gk_up[ik] @ self.Delta @ Gk_dn[ikq]
                    #for ik, ikq in enumerate(range(self.nkpts)):
                    Guk = Gk_up[ik, :, :]
                    #Gdk = Gk_dn[ik, :, :]
                    #Gukq = Gk_up[ikq, :, :]
                    Gdkq = Gk_dn[ikq, :, :]
                    for i, iatom in enumerate(self.ind_mag_atoms):
                        Deltai = self.get_Delta(iatom)
                        for j, jatom in enumerate(self.ind_mag_atoms):
                            Deltaj = self.get_Delta(jatom)
                            Gij_up_k = self.Gk_atom(Guk, iatom, jatom)
                            Gji_dn_kq = self.Gk_atom(Gdkq, jatom, iatom)

                            Gij_dn_kq = self.Gk_atom(Gdk, iatom, jatom)
                            Gji_up_k = self.Gk_atom(Gukq, jatom, iatom)
                            A = np.trace(
                                np.linalg.multi_dot(
                                    (Deltai, Gij_up_k, Deltaj, Gji_dn_kq)))
                            K1 = Gij_up_k @ Deltaj @ Gji_dn_kq
                
                            K2 = Gij_dn_kq @ Deltaj @ Gji_up_k
                            X = Gij_up_k @ Gji_dn_kq
                            A = (np.trace(Deltai @ K1) +
                                 np.trace(Deltai @ K2)) * 0.5
                            K = K1 + K2
                            K += Gij_dn_kq @ Deltaj @ Gji_up_k
                            #K=Kq[self.orb_slice[iatom], self.orb_slice[iatom]]*2
                            #self.Jqe_orb_list[(ikq, iatom, jatom)].append(tmp / (4.0 * np.pi))
                            self.Jqe_list[ie, iq, i,
                                          j] += A / (4.0 * np.pi) / self.nqpt
                            self.Kqe_list[
                                ie, iq, i,
                                j] -= np.trace(K) / (2.0 * np.pi) / self.nqpt
                            self.Xqe_list[
                                ie, iq, i,
                                j] += np.trace(X) * (2.0 / np.pi) / self.nqpt

    def integrate(self, method="simpson"):
        self.Jq = np.zeros((self.nqpt, self.nmagatom, self.nmagatom),
                           dtype=float)
        self.Kq = np.zeros_like(self.Jq)
        self.Xq = np.zeros_like(self.Jq)
        self.Jnorm_q = np.zeros_like(self.Jq)
        if method == "trapezoidal":
            integrate = trapezoidal_nonuniform
        elif method == 'simpson':
            integrate = simpson_nonuniform
        #self.rho_up = np.imag(integrate(self.contour.path, self.rho_up_list))
        #self.rho_dn = np.imag(integrate(self.contour.path, self.rho_dn_list))
        for iq, q in enumerate(self.qpts):
            for i, iatom in enumerate(self.ind_mag_atoms):
                for j, jatom in enumerate(self.ind_mag_atoms):
                    self.Jq[iq, i, j] = np.imag(
                        integrate(self.contour.path, self.Jqe_list[:, iq, i,
                                                                   j]))
                    self.Kq[iq, i, j] = np.imag(
                        integrate(self.contour.path, self.Kqe_list[:, iq, i,
                                                                   j]))
                    self.Xq[iq, i, j] = np.imag(
                        integrate(self.contour.path, self.Xqe_list[:, iq, i,
                                                                   j]))
            #print(f"{q=}, {self.Kq[iq]}")
        return self.Jq

    def bruno_renormalize(self):
        M = np.diag(self.spinat[self.ind_mag_atoms, 2])

        for iq, q in enumerate(self.qpts):
            self.Jnorm_q[iq] = self.Jq[iq] + 0.5 * (
                M - self.Kq[iq].T) @ np.linalg.inv(
                    self.Xq[iq]) @ (M - self.Kq[iq])
            #print(f"{q}: Jq:{self.Jq[iq]}, Jnorm:{self.Jnorm_q[iq]}")

    def q_to_r(self):
        self.JR = np.zeros((len(self.Rlist), self.nmagatom, self.nmagatom),
                           dtype=float)
        self.Jnorm_R = np.zeros_like(self.JR)
        self.KR = np.zeros_like(self.JR)
        for iR, R in enumerate(self.Rlist):
            for iq, q in enumerate(self.qpts):
                phase = np.exp(-2.0j * np.pi * (R @ q))
                self.JR[iR] += np.real(self.Jq[iq] * phase) / len(self.qpts)
                self.Jnorm_R[iR] += np.real(self.Jnorm_q[iq] * phase) / len(
                    self.qpts)
                self.KR[iR] += np.real(self.Kq[iq] * phase) / len(self.qpts)
            #print(f"{R}: J={self.JR[iR]}, Jnorm{self.Jnorm_R[iR]}")
            #print(f"{R}: {np.sum(self.KR, axis=0)}")
        return self.JR

    def get_Jdict(self):
        for iR, R in enumerate(self.Rlist):
            R = tuple(R)
            for i, iatom in enumerate(self.ind_mag_atoms):
                for j, jatom in enumerate(self.ind_mag_atoms):
                    val = self.JR[iR, i, j]
                    ispin = self.ispin(iatom)
                    jspin = self.ispin(jatom)
                    keyspin = (R, ispin, jspin)
                    is_nonself = not (R == (0, 0, 0) and iatom == jatom)
                    Jij = val / np.sign(
                        np.dot(self.spinat[iatom], self.spinat[jatom]))
                    #Jorbij = np.imag(self.Jorb[key]) / np.sign(
                    #    np.dot(self.spinat[iatom], self.spinat[jatom]))
                    if is_nonself:
                        self.exchange_Jdict[keyspin] = Jij
                        #self.exchange_Jdict_orb[keyspin] = Jorbij

    def calculate_all(self):
        self._prepare()
        self.get_all_A()
        self.integrate()
        self.bruno_renormalize()
        self.q_to_r()
        self.get_Jdict()



import re
from collections import defaultdict
import numpy as np


def split_orb_name(name):
    """
    split name to : n, l, label
    """
    m = re.findall(r"(\d[a-z\d\-]*)(.*)", name)
    m = m[0]
    return m[0], m[1]


def map_orbs_matrix(orblist, spinor=False, include_only=None):
    if spinor:
        orblist = orblist[::2]

    norb = len(orblist)

    ss = [split_orb_name(orb) for orb in orblist]
    orbdict = dict(zip(ss, range(norb)))

    reduced_orbdict = defaultdict(lambda: [])

    if include_only is None:
        for key, val in orbdict.items():
            reduced_orbdict[key[0]].append(val)
    else:
        for key, val in orbdict.items():
            if key[0][:2] in include_only:
                reduced_orbdict[key[0]].append(val)

    reduced_orbs = tuple(reduced_orbdict.keys())
    ngroup = len(reduced_orbdict)
    mmat = np.zeros((norb, ngroup), dtype=int)

    for i, (key, val) in enumerate(reduced_orbdict.items()):
        for j in val:
            mmat[j, i] = 1
    return mmat, reduced_orbs


def test_split():
    split_orb_name('3sZ1')
    split_orb_name('3dxyZ1')
    split_orb_name('5dxyZ1')
    split_orb_name('5dx2-y2Z1P')


def test():
    odict = {0: ['3sZ1', '3sZ1', '4sZ1', '4sZ1', '4sZ2', '4sZ2', '3pyZ1', '3pyZ1', '3pzZ1', '3pzZ1', '3pxZ1', '3pxZ1', '3dxyZ1', '3dxyZ1', '3dyzZ1', '3dyzZ1', '3dz2Z1', '3dz2Z1', '3dxzZ1', '3dxzZ1', '3dx2-y2Z1', '3dx2-y2Z1', '3dxyZ2', '3dxyZ2', '3dyzZ2', '3dyzZ2', '3dz2Z2', '3dz2Z2', '3dxzZ2', '3dxzZ2', '3dx2-y2Z2', '3dx2-y2Z2', '4pyZ1P', '4pyZ1P', '4pzZ1P', '4pzZ1P', '4pxZ1P', '4pxZ1P'], 1: ['3sZ1', '3sZ1', '4sZ1', '4sZ1', '4sZ2', '4sZ2', '3pyZ1', '3pyZ1', '3pzZ1', '3pzZ1', '3pxZ1', '3pxZ1', '3dxyZ1', '3dxyZ1', '3dyzZ1', '3dyzZ1', '3dz2Z1', '3dz2Z1', '3dxzZ1', '3dxzZ1', '3dx2-y2Z1', '3dx2-y2Z1', '3dxyZ2', '3dxyZ2', '3dyzZ2', '3dyzZ2', '3dz2Z2', '3dz2Z2', '3dxzZ2', '3dxzZ2', '3dx2-y2Z2', '3dx2-y2Z2', '4pyZ1P', '4pyZ1P', '4pzZ1P', '4pzZ1P', '4pxZ1P', '4pxZ1P'], 2: ['5sZ1', '5sZ1', '5sZ2', '5sZ2', '5pyZ1', '5pyZ1', '5pzZ1', '5pzZ1', '5pxZ1', '5pxZ1', '5pyZ2', '5pyZ2', '5pzZ2', '5pzZ2', '5pxZ2', '5pxZ2', '5dxyZ1P', '5dxyZ1P', '5dyzZ1P', '5dyzZ1P', '5dz2Z1P', '5dz2Z1P', '5dxzZ1P', '5dxzZ1P', '5dx2-y2Z1P', '5dx2-y2Z1P'], 3: ['5sZ1', '5sZ1', '5sZ2', '5sZ2', '5pyZ1', '5pyZ1', '5pzZ1', '5pzZ1', '5pxZ1', '5pxZ1', '5pyZ2', '5pyZ2', '5pzZ2', '5pzZ2', '5pxZ2',
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            '5pxZ2', '5dxyZ1P', '5dxyZ1P', '5dyzZ1P', '5dyzZ1P', '5dz2Z1P', '5dz2Z1P', '5dxzZ1P', '5dxzZ1P', '5dx2-y2Z1P', '5dx2-y2Z1P'], 4: ['5sZ1', '5sZ1', '5sZ2', '5sZ2', '5pyZ1', '5pyZ1', '5pzZ1', '5pzZ1', '5pxZ1', '5pxZ1', '5pyZ2', '5pyZ2', '5pzZ2', '5pzZ2', '5pxZ2', '5pxZ2', '5dxyZ1P', '5dxyZ1P', '5dyzZ1P', '5dyzZ1P', '5dz2Z1P', '5dz2Z1P', '5dxzZ1P', '5dxzZ1P', '5dx2-y2Z1P', '5dx2-y2Z1P'], 5: ['5sZ1', '5sZ1', '5sZ2', '5sZ2', '5pyZ1', '5pyZ1', '5pzZ1', '5pzZ1', '5pxZ1', '5pxZ1', '5pyZ2', '5pyZ2', '5pzZ2', '5pzZ2', '5pxZ2', '5pxZ2', '5dxyZ1P', '5dxyZ1P', '5dyzZ1P', '5dyzZ1P', '5dz2Z1P', '5dz2Z1P', '5dxzZ1P', '5dxzZ1P', '5dx2-y2Z1P', '5dx2-y2Z1P'], 6: ['5sZ1', '5sZ1', '5sZ2', '5sZ2', '5pyZ1', '5pyZ1', '5pzZ1', '5pzZ1', '5pxZ1', '5pxZ1', '5pyZ2', '5pyZ2', '5pzZ2', '5pzZ2', '5pxZ2', '5pxZ2', '5dxyZ1P', '5dxyZ1P', '5dyzZ1P', '5dyzZ1P', '5dz2Z1P', '5dz2Z1P', '5dxzZ1P', '5dxzZ1P', '5dx2-y2Z1P', '5dx2-y2Z1P'], 7: ['5sZ1', '5sZ1', '5sZ2', '5sZ2', '5pyZ1', '5pyZ1', '5pzZ1', '5pzZ1', '5pxZ1', '5pxZ1', '5pyZ2', '5pyZ2', '5pzZ2', '5pzZ2', '5pxZ2', '5pxZ2', '5dxyZ1P', '5dxyZ1P', '5dyzZ1P', '5dyzZ1P', '5dz2Z1P', '5dz2Z1P', '5dxzZ1P', '5dxzZ1P', '5dx2-y2Z1P', '5dx2-y2Z1P']}

    olist = odict[0]
    map_orbs_matrix(olist, spinor=True)
    map_orbs_matrix(olist, spinor=True, include_only=['3d'])


if __name__=="__main__":
    test()



try:
    from gpaw.lcao.tightbinding import TightBinding
    from gpaw import restart, GPAW
    from gpaw.lcao.tools import get_lead_lcao_hamiltonian, get_lcao_hamiltonian, get_bf_centers, get_bfi
    _has_gpaw=True
except:
    _has_gpaw=False    

#from banddownfolder.scdm.downfolder import BandDownfolder
import numpy as np
from scipy.linalg import eigh
from ase.dft.kpoints import monkhorst_pack
from ase.units import Ha
import pickle

class GPAWWrapper():
    """
    https://gitlab.com/gpaw/gpaw/-/blob/master/gpaw/lcao/tightbinding.py
    """
    def __init__(self, atoms=None, calc=None, gpw_fname=None):
        if not _has_gpaw:
            raise ImportError("GPAW was not imported. Please install gpaw.")
        if atoms is None or calc is None:
            atoms, calc=restart(gpw_fname, fixdensity=False,  txt='nscf.txt')
        self.atoms=atoms
        self.calc=calc
        #self.wfs=self.calc.wfs
        #self.h=self.calc.hamiltonian
        #self.kpt_u = self.wfs.kpt_u
        self.cell = self.atoms.get_cell()
        #self.wann_centers=get_bf_centers(self.atoms)
        self.positions=self.cell.scaled_positions(get_bf_centers(self.atoms))
        #if calc.get_spin_polarized():
        #    self.positions=np.vstack([self.positions, self.positions])
        self.efermi=self.calc.get_fermi_level()
        self.R2kfactor= -2.0j* np.pi
        #self.nbasis=len(self.positions)
        self.norb=len(self.positions)
        self.nbasis=self.norb*2
        self._name='GPAW'


    #def gen_ham(self, k):
    #    H_MM=self.wfs.eigensolver.calculate_hamiltonian_matrix(self.h, self.wfs, k)
    #    tri2full(H_MM*Ha)
    #    return H_MM

    #def Sk(self, k):
    #    S_MM=self.wfs.S_qMM[kpt.q]
    #    tri2full(S_MM)
    #    return S_MM


    #def solve(self, k):
    #    return eigh(self.hamk(k), self.S(k))

    def get_kpts(self):
        return calc.get_ibz_k_points()

    def HS_and_eigen(self, kpts=None, convention=2):
        if kpts is not None:
            self.calc.set(kpts=kpts)
            self.atoms.calc=self.calc
            self.atoms.get_total_energy()

        self.calc.set(symmetry='off', fixdensity=True)
        self.atoms.get_total_energy()
        #self.kpt_u = self.wfs.kpt_u
        H, S= get_lcao_hamiltonian(self.calc)
        np.save('kpts.npy', kpts)
        np.save('H.npy', H)
        np.save('S.npy', S)

        nspin, nkpt, norb, _ = H.shape
        nbasis=nspin*norb
        evals=np.zeros((nkpt, nbasis), dtype=float)
        evecs=np.zeros((nkpt, nbasis, nbasis),dtype=complex)
        if self.calc.get_spin_polarized():
            H2=np.zeros((nkpt, nbasis, nbasis),dtype=complex)
            # spin up
            H2[:, :norb, :norb]=H[0]
            # spin down
            H2[:, norb:, norb:]=H[1]
            S2=np.zeros((nkpt, nbasis, nbasis),dtype=complex)
            S2[:, :norb, :norb]=S
            S2[:, norb:, norb:]=S

            for ikpt, k in enumerate(self.calc.get_ibz_k_points()):
                evals0, evecs0 = eigh(H[0, ikpt,:,:], S[ikpt,:,:])
                evals1, evecs1 = eigh(H[1, ikpt,:,:], S[ikpt,:,:])
                evals[ikpt, :norb]=evals0
                evals[ikpt, norb:]=evals1
                evecs[ikpt, :norb, :norb]=evecs0
                evecs[ikpt, norb:, norb:]=evecs1
        else:
            H2=H[0]
            for ikpt, k in self.kpt_u:
                evals[ikpt], evecs[ikpt]= eigh(H[0, ikpt], S[ikpt])
        np.save('kpts.npy', kpts)
        np.save('H2.npy', H2)
        np.save('S2.npy', S2)
        np.save('evals.npy', evals)
        np.save('evecs.npy', evecs)
        return H2, S2, evals, evecs

    def solve_all(self, kpts, convention=2):
        H, evals, evecs=self.H_and_eigen(kpts=kpts, convention=convention)
        return evals, evecs



class GPAWTBWrapper():
    def __init__(self, calc=None, atoms=None,  gpw_fname=None, pickle_fname=None):
        if not _has_gpaw:
            raise ImportError("GPAW was not imported. Please install gpaw.")
        if pickle_fname is None:
            if gpw_fname is not None:
                calc=GPAW(gpw_fname, fixdensity=True, symmetry='off', txt='TB2J_wrapper.log', basis='dzp', mode='lcao')
                atoms=calc.atoms
                atoms.calc=calc
                E = atoms.get_potential_energy()
            tb=TightBinding(atoms, calc)
            self.H_NMM, self.S_NMM=tb.h_and_s()
            self.Rlist=tb.R_cN.T
        else:
            with open(fname, 'rb') as myfile:
                self.H_NMM, self.S_NMM, self.Rlist=pickle.load(myfile)
        self.nR, self.nbasis, _ = self.H_NMM.shape
        self.positions=np.zeros((self.nbasis, 3))

    def save_pickle(self, fname):
        with open(fname, 'wb') as myfile:
            pickle.dump([self.H_NMM, self.S_NMM, self.Rlist], myfile)

    def solve(self, k):
        self.Hk=np.zeros((self.nbasis, self.nbasis), dtype='complex')
        self.Sk=np.zeros((self.nbasis, self.nbasis), dtype='complex')
        for iR, R in enumerate(self.Rlist):
            phase=np.exp(-2j*np.pi* np.dot(k, R))
            self.Hk+=self.H_NMM[iR]*phase
            self.Sk+=self.S_NMM[iR]*phase
        self.Sk=np.real((self.Sk+self.Sk.T)/2)
        return eigh(self.Hk, self.Sk)
        try:
            return eigh(self.Hk, self.Sk)
        except:
            return np.zeros(self.nbasis), np.zeros((self.nbasis, self.nbasis))

    def solve_all(self, kpts):
        evals = []
        evecs = []
        for ik, k in enumerate(kpts):
            evalue, evec = self.solve(k)
            evals.append(evalue)
            evecs.append(evec)
        return np.array(evals, dtype=float), np.array(evecs,
                                                      dtype=complex,
                                                      order='C')



def test_Ham():
    calc=GPAW('/home/hexu/projects/gpaw/bccFe/atoms_nscf.gpw', fixdensity=True, symmetry='off',  txt='nscf.txt')
    atoms=calc.atoms
    atoms.calc=calc
    calc.set(kpts=(3,3,3))
    E = atoms.get_potential_energy()
    tb=TightBinding(atoms, calc)
    kpath=monkhorst_pack([3,3,3])
    evals, evecs=tb.band_structure(kpath, blochstates=True)


def test():
    g=GPAWWrapper(fname='STO_nscf.gpw')
    g.H_and_eigen(kpts=monkhorst_pack(2,2,2))
    #g.save_pickle('model.pickle')
    #df=GPAWDownfolder('model.pickle')
    #df.downfold(nwann=1, anchor_kpt=[0,0,0] )
    #df.plot_band_fitting(npoints=100)

#test_Ham()

def test():
    g=GPAWWrapper(gpw_fname='/home/hexu/projects/gpaw/bccFe/atoms_nscf.gpw')
    H, E, V=g.H_and_eigen(kpts=monkhorst_pack([2,2,2]))
    print(H.shape)
    print(E.shape)
    print(E)
    print(V.shape)





import os
import pickle
import copy
from collections import defaultdict
import numpy as np
from ase.dft.kpoints import monkhorst_pack
from TB2J.io_exchange import SpinIO


class JDownfolder():
    def __init__(self, JR, Rlist, iM, iL, qmesh):
        self.JR = JR
        self.Rlist = Rlist
        self.nR = len(Rlist)
        self.nM = len(iM)
        self.nL = len(iL)
        self.nsite = self.nM + self.nL
        self.iM = iM
        self.iL = iL
        self.qmesh = qmesh
        self.qpts = monkhorst_pack(qmesh)
        self.nqpt = len(self.qpts)

    def get_Jq(self, q):
        Jq = np.zeros(self.JR[0].shape, dtype=complex)
        for iR, R in enumerate(self.Rlist):
            phase = np.exp(2.0j * np.pi * np.dot(q, R))
            Jq += self.JR[iR] * phase
        return Jq

    def get_JR(self):
        JR_downfolded = np.zeros((self.nR, self.nM, self.nM), dtype=float)
        Jq_downfolded = np.zeros((self.nqpt, self.nM, self.nM), dtype=complex)
        for iq, q in enumerate(self.qpts):
            Jq = self.get_Jq(q)
            Jq_downfolded[iq] = self.downfold_oneq(Jq)

            for iR, R in enumerate(self.Rlist):
                phase = np.exp(-2.0j * np.pi * np.dot(q, R))
                JR_downfolded[iR] += np.real(Jq_downfolded[iq] * phase /
                                             self.nqpt)
        return JR_downfolded

    def downfold_oneq(self, J):
        JMM = J[np.ix_(self.iM, self.iM)]
        JLL = J[np.ix_(self.iL, self.iL)]
        JLM = J[np.ix_(self.iL, self.iM)]
        JML = J[np.ix_(self.iM, self.iL)]
        Jn = JMM - JML @ np.linalg.inv(JLL) @ JLM
        return Jn


class JDownfolder_pickle():
    def __init__(self, inpath, metals, ligands, outpath):
        fname=os.path.join(inpath, 'TB2J.pickle')
        with open(fname, 'rb') as myfile:
            self.obj = pickle.load(myfile)

        self.atoms = self.obj['atoms']
        atoms = self.atoms
        index_spin = self.obj['index_spin']
        ind_atoms = self.obj['ind_atoms']
        Jdict = self.obj['exchange_Jdict']
        nspin = len([x for x in index_spin if x>=0])


        self.natom=len(self.atoms)

        self.magnetic_elements= metals
        self.iM=[]
        self.iL=[]
        self.ind_mag_atoms=[]
        for i, sym in enumerate(self.atoms.get_chemical_symbols()):
            if sym in self.magnetic_elements:
                self.iM.append(index_spin[i])
                self.ind_mag_atoms.append(i)
            elif sym in ligands:
                self.iL.append(index_spin[i])

        self.nM=len(self.iM)
        self.nL=len(self.iL)
        self.nsite=self.nM+self.nL


        # build matrix form of JR
        JR = defaultdict(lambda: np.zeros((self.nsite, self.nsite), dtype=float))
        for key, val in Jdict.items():
            R, i, j = key
            JR[R][i, j] = val

        Rlist = list(JR.keys())

        nR = len(Rlist)
        JR2 = np.zeros((nR, self.nsite, self.nsite), dtype=float)
        for i, (key, val) in enumerate(JR.items()):
            JR2[i] = val


        # sum rule
        iR0 = np.argmin(np.linalg.norm(Rlist, axis=1))
        for i in range(self.nsite):
            sum_JRi=np.sum(np.sum(JR2, axis=0)[i])
            JR2[iR0][i, i] -= sum_JRi

        d = JDownfolder(JR2, Rlist, iM=self.iM, iL=self.iL, qmesh=[7, 7, 7])
        Jd = d.get_JR()

        self.Rlist=Rlist
        self.Rcut=None

        self._prepare_distance()
        self._prepare_index_spin()
        self.Jdict={}
        for iR, R in enumerate(d.Rlist):
            for i, ispin in enumerate(self.index_spin):
                for j, jspin in enumerate(self.index_spin):
                    if ispin>=0 and jspin>=0:
                        if not (tuple(R) ==(0,0,0) and ispin==jspin): 
                            # self interaction.
                            self.Jdict[(tuple(R), ispin, jspin)]=Jd[iR, ispin, jspin]


        io = SpinIO(atoms=atoms,
                    spinat=self.obj['spinat'],
                    charges=self.obj['charges'],
                    index_spin=self.index_spin,
                    colinear=True,
                    distance_dict=self.distance_dict,
                    exchange_Jdict=self.Jdict)

        io.write_all(outpath)

    def _prepare_distance(self):
        self.distance_dict = {}
        self.short_Rlist = []
        self.R_ijatom_dict = defaultdict(lambda: [])
        ind_matoms = self.ind_mag_atoms
        for ispin, iatom in enumerate(ind_matoms):
            for jspin, jatom in enumerate(ind_matoms):
                for R in self.Rlist:
                    pos_i = self.atoms.get_positions()[iatom]
                    pos_jR = self.atoms.get_positions()[jatom] + np.dot(
                        R, self.atoms.get_cell())
                    vec = pos_jR - pos_i
                    distance = np.sqrt(np.sum(vec**2))
                    if self.Rcut is None or distance < self.Rcut:
                        self.distance_dict[(tuple(R), ispin,
                                            jspin)] = (vec, distance)
                        self.R_ijatom_dict[tuple(R)].append((iatom, jatom))
        self.short_Rlist = list(self.R_ijatom_dict.keys())

    def _prepare_index_spin(self):
        # index_spin: index in spin hamiltonian of atom. starts from 1. -1 means not considered.
        ind_matoms = []
        self.index_spin = []
        ispin = 0
        for i, sym in enumerate(self.atoms.get_chemical_symbols()):
            if sym in self.magnetic_elements:
                ind_matoms.append(i)
                self.index_spin.append(ispin)
                ispin += 1
            else:
                self.index_spin.append(-1)




def test(inpath, outpath, metals, ligands):
    #path = "/home/hexu/projects/NiCl2/vasp_inputs/TB2J_results"
    path = "/home/hexu/projects/NiCl2/AFM/TB2J_results"
    #fname = os.path.join(path, "TB2J.pickle")
    p=JDownfolder_pickle(inpath=inpath, metals=['Ni'], ligands=['Cl'], outpath='TB2J_results_downfolded')


if __name__=="__main__":
    test()



import numpy as np
from ase.atoms import Atoms

class BaseSpinModelParser(object):
    """
    SpinModelParser: a general model for spin model file parser.
    """

    def __init__(self, fname):
        self._fname = fname
        self.atoms = None
        self.damping_factors = []
        self.gyro_ratios = []
        self.index_spin = []
        self.cell = None
        self.zions = []
        self.masses = []
        self.positions = []
        self.spinat = []
        self._exchange = {}
        self._dmi = {}
        self._bilinear = {}
        self._parse(fname)
        self.lattice = Atoms(
            positions=self.positions, masses=self.masses, cell=self.cell)

    def _parse(self, fname):
        raise NotImplementedError("parse function not implemented yet")

    def get_atoms(self):
        return self.atoms

    def _spin_property(self, prop):
        return [
            prop[i] for i in range(len(self.index_spin))
            if self.index_spin[i] > 0
        ]

    @property
    def spin_positions(self):
        return np.array(self._spin_property(self.positions), dtype='float')

    @property
    def spin_zions(self):
        return np.array(self._spin_property(self.zions), dtype='int')

    @property
    def spin_spinat(self):
        return np.array(self._spin_property(self.spinat), dtype='float')

    @property
    def spin_damping_factors(self):
        return np.array(
            self._spin_property(self.damping_factors), dtype='float')

    @property
    def spin_gyro_ratios(self):
        return np.array(self._spin_property(self.gyro_ratios), dtype='float')

    def get_index_spin(self):
        return self.index_spin

    def exchange(self, isotropic=True):
        if isotropic:
            iso_jdict = {}
            for key, val in self._exchange.items():
                iso_jdict[key] = val[0]
            return iso_jdict
        else:
            return self._exchange

    @property
    def dmi(self):
        return self._dmi

    @property
    def has_exchange(self):
        return bool(len(self._exchange))

    @property
    def has_dmi(self):
        return bool(len(self._dmi))

    @property
    def has_bilinear(self):
        return bool(len(self._bilinear))




"""
Module about Pauli (and I) matrices.
"""

import numpy as np
from numpy import zeros_like
from scipy.linalg import svd

s0 = np.array([[1, 0], [0, 1]])
s1 = np.array([[0, 1], [1, 0]])
s2 = np.array([[0, -1j], [1j, 0]])
s3 = np.array([[1, 0], [0, -1]])

s0T = s0.T
s1T = s1.T
s2T = s2.T
s3T = s3.T

pauli_dict = {0: s0, 1: s1, 2: s2, 3: s3}


def pauli_mat(nbasis, i):
    """
    nbasis: size of the matrix. should be multiple of 2.
    i: index of pauli dictionary.
    """
    N = nbasis // 2
    assert (N * 2 == nbasis)
    M = np.ones((N, N), dtype='complex')
    spm = pauli_dict[i]
    return np.block([[M * spm[0, 0], M * spm[0, 1]],
                     [M * spm[1, 0], M * spm[1, 1]]])


def pauli_decomp(M):
    """ Given a 2*2 matrix, get the I, x, y, z component.
    :param M: 2*2 matrix
    :returns:  (I, x, y, z) are four scalars.
    :rtype: same as dtype of M
    """
    return (np.trace(s0.dot(M)) / 2, np.trace(s1.dot(M)) / 2,
            np.trace(s2.dot(M)) / 2, np.trace(s3.dot(M)) / 2)


def pauli_decomp2(M):
    """ Given a 2*2 matrix, get the I, x, y, z component. (method2)
    :param M: 2*2 matrix
    :returns:  (I, x, y, z) are four scalars.
    :rtype: same as dtype of M
    """
    return (np.sum(M * s0T) / 2, np.sum(M * s1T) / 2, np.sum(M * s2T) / 2,
            np.sum(M * s3T) / 2)


def pauli_sigma_norm(M):
    MI, Mx, My, Mz = pauli_decomp2(M)
    return np.linalg.norm([Mx, My, Mz])


def pauli_block_I(M, norb):
    """
    I compoenent of a matrix, see pauli_block
    """
    ret = zeros_like(M)
    tmp = (M[::2, ::2] + M[1::2, 1::2]) / 2
    ret[::2, ::2] = ret[1::2, 1::2] = tmp
    return ret


def pauli_block_x(M, norb):
    """
    x compoenent of a matrix, see pauli_block
    """
    ret = zeros_like(M)
    tmp = (M[::2, 1::2] + M[1::2, ::2]) / 2
    ret[::2, 1::2] = ret[1::2, ::2] = tmp
    return ret


def pauli_block_y(M, norb):
    """
    y compoenent of a matrix, see pauli_block
    """
    ret = zeros_like(M)
    tmp = (M[::2, 1::2] * (-1j) + M[1::2, ::2] * (1j)) / 2
    ret[::2, 1::2] = tmp * (-1j)
    ret[1::2, ::2] = tmp * 1j
    return tmp, ret


def pauli_block_z(M, norb):
    """ z compoenent of a matrix, see pauli_block
    :param M:
    :param norb:
    :returns:
    :rtype:
    """
    ret = zeros_like(M)
    tmp = (M[::2, ::2] - M[1::2, 1::2]) / 2
    ret[::2, ::2] = tmp
    ret[1::2, 1::2] = -tmp
    return tmp, ret


def pauli_block(M, idim):
    """ Get the I, x, y, z component of a matrix.
    :param M: The input matrix,  aranged in four blocks:
    [[upup, updn], [dnup, dndn]]. let norb be number of orbitals in
    each block. (so M has dim 2norb1*2norb2)
    :param idim: 0, 1,2, 3 for I , x, y, z.
    :returns:  the idim(th) component of M
    :rtype: a matrix with shape of M.shape//2
    """
    # ret = zeros_like(M)
    if idim == 0:
        tmp = (M[::2, ::2] + M[1::2, 1::2]) / 2.0
    elif idim == 1:
        tmp = (M[::2, 1::2] + M[1::2, ::2]) / 2.0
    elif idim == 2:
        # Note that this is not element wise product with sigma_y but dot product
        # sigma_y=[[0, -1j],[1j, 0]]
        tmp = (M[::2, 1::2] * (1.0j) + M[1::2, ::2] * (-1.0j)) / 2.0
    elif idim == 3:
        tmp = (M[::2, ::2] - M[1::2, 1::2]) / 2.0
    else:
        raise NotImplementedError()
    return tmp


def pauli_block_all(M):
    MI = (M[::2, ::2] + M[1::2, 1::2]) / 2.0
    Mx = (M[::2, 1::2] + M[1::2, ::2]) / 2.0
    # Note that this is not element wise product with sigma_y but dot product
    My = (M[::2, 1::2] - M[1::2, ::2]) * 0.5j
    Mz = (M[::2, ::2] - M[1::2, 1::2]) / 2.0
    return MI, Mx, My, Mz


def op_norm(M):
    return max(svd(M)[1])


def pauli_block_sigma_norm(M):
    """
    M= MI * I + \vec{P} dot \vec{sigma}
    = MI*I + p * \vec{e} dot \vec{sigma}
    where p is the norm of P.
    """
    MI, Mx, My, Mz = pauli_block_all(M)
    ex, ey, ez = np.trace(Mx), np.trace(My), np.trace(Mz)
    #ex,ey,ez = op_norm(Mx), op_norm(My), op_norm(Mz)
    evec = np.array((ex, ey, ez))
    evec = evec / np.linalg.norm(evec)
    return Mx * evec[0] + My * evec[1] + Mz * evec[2]



import os
import shutil
import numpy as np
from ase.atoms import Atoms
from TB2J.utils import symbol_number
from collections import defaultdict
from scipy.linalg import eigh
from TB2J.myTB import AbstractTB


class SislWrapper(AbstractTB):
    def __init__(self, sisl_hamiltonian, spin=None):
        self.is_siesta = False
        self.is_orthogonal = False
        self.ham = sisl_hamiltonian
        # k2Rfactor : H(k) = \int_R H(R) * e^(k2Rfactor * k.R)
        self.R2kfactor = 2.0j * np.pi  #
        if spin == 'up':
            spin = 0
        elif spin == 'down':
            spin = 1
        if spin not in [None, 0, 1, 'merge']:
            raise ValueError("spin should be None/0/1, but is %s" % spin)
        self.spin = spin
        self.orbs = []
        self.orb_dict = defaultdict(lambda: [])
        g = self.ham._geometry
        _atoms = self.ham._geometry._atoms
        atomic_numbers = []
        self.positions = g.xyz
        self.cell = np.array(g.sc.cell)
        for ia, a in enumerate(_atoms):
            atomic_numbers.append(a.Z)
        self.atoms = Atoms(numbers=atomic_numbers,
                           cell=self.cell,
                           positions=self.positions)
        sdict = list(symbol_number(self.atoms).keys())
        if self.ham.spin.is_colinear and (self.spin in [0, 1]):
            for ia, a in enumerate(_atoms):
                symnum = sdict[ia]
                try:
                    orb_names = [f"{symnum}|{x.name()}|up" for x in a.orbital]
                except:
                    orb_names = [f"{symnum}|{x.name()}|up" for x in a.orbitals]
                self.orbs += orb_names
                self.orb_dict[ia] += orb_names
            self.norb = len(self.orbs)
            self.nbasis = self.norb
        elif self.ham.spin.is_spinorbit or self.ham.spin.is_noncolinear or self.spin == 'merge':
            for ia, a in enumerate(_atoms):
                symnum = sdict[ia]
                orb_names = []
                for x in a.orbitals: 
                    for spin in {'up', 'down'}:
                        orb_names.append(f"{symnum}|{x.name()}|{spin}")
                self.orbs += orb_names
                self.orb_dict[ia] += orb_names
            self.norb = len(self.orbs) // 2
            self.nbasis = len(self.orbs)
        else:
            raise ValueError(
                "The hamiltonian should be either spin-orbit or colinear")
        self._name = 'SIESTA'


    def view_info(self):
        print(self.orb_dict)
        print(self.atoms)

    def solve(self, k, convention=2):
        if convention == 1:
            gauge = 'r'
        elif convention == 2:
            gauge = 'R'
        if self.spin in [0, 1]:
            evals, evecs = self.ham.eigh(k=k,
                                         spin=self.spin,
                                         eigvals_only=False,
                                         gauge=gauge)
        elif self.spin is None:
            evals, evecs = self.ham.eigh(k=k, eigvals_only=False, gauge=gauge)
        elif self.spin == 'merge':
            evals0, evecs0 = self.ham.eigh(k=k,
                                           spin=0,
                                           eigvals_only=False,
                                           gauge=gauge)
            evals1, evecs1 = self.ham.eigh(k=k,
                                           spin=1,
                                           eigvals_only=False,
                                           gauge=gauge)
            evals = np.zeros(self.nbasis, dtype=float)
            evecs = np.zeros((self.nbasis, self.nbasis), dtype=complex)
            evals[::2] = evals0
            evals[1::2] = evals1
            evecs[::2, ::2] = evecs0
            evecs[1::2, 1::2] = evecs1

        return evals, evecs

    def Hk(self, k, convention=2):
        if convention == 1:
            gauge = 'r'
        elif convention == 2:
            gauge = 'R'
        if self.spin is None:
            H = self.ham.Hk(k, gauge=gauge, format='dense')
        elif self.spin in [0, 1]:
            H = self.ham.Hk(k, spin=self.spin, gauge=gauge, format='dense')
        elif self.spin == 'merge':
            H = np.zeros((self.nbasis, self.nbasis), dtype='complex')
            H[::2, ::2] = self.ham.Hk(k,
                                                    spin=0,
                                                    gauge=gauge,
                                                    format='dense')
            H[1::2, 1::2] = self.ham.Hk(k,
                                                    spin=1,
                                                    gauge=gauge,
                                                    format='dense')
        return H

    def eigen(self, k, convention=2):
        return self.solve(k)

    def gen_ham(self, k, convention=2):
        return self.Hk(k, convention=convention)

    def Sk(self, k, convention=2):
        if convention == 1:
            gauge = 'r'
        elif convention == 2:
            gauge = 'R'
        S0 = self.ham.Sk(k, gauge='R', format='dense')
        if self.spin is None:
            S = S0
        elif self.spin in [0, 1]:
            S = S0
        elif self.spin == 'merge':
            S = np.zeros((self.nbasis, self.nbasis), dtype='complex')
            S[::2, ::2] = S0
            S[1::2, 1::2] = S0
        return S

    def solve_all(self, kpts, orth=False):
        evals = []
        evecs = []
        for ik, k in enumerate(kpts):
            if orth:
                S = self.Sk(k)
                Smh = Lowdin(S)
                H = self.gen_ham(k)
                Horth = Smh.T.conj() @ H @ Smh
                evalue, evec = eigh(Horth)
            else:
                evalue, evec = self.solve(k)
            evals.append(evalue)
            evecs.append(evec)
        return np.array(evals, dtype=float), np.array(evecs,
                                                      dtype=complex,
                                                      order='C')

    def HSE_k(self, k,convention=2):
        Hk = self.Hk(k, convention=convention)
        Sk = self.Sk(k, convention=convention)
        evalue, evec = self.solve(k, convention=convention)
        return Hk, Sk, evalue, evec


    def HS_and_eigen(self, kpts, convention=2):
        nkpts = len(kpts)
        evals = np.zeros((nkpts, self.nbasis), dtype=float)
        self.nkpts=nkpts
        if not self._use_cache:
            evecs = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
            H = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
            S = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
        else:
            self._prepare_cache()

        for ik, k in enumerate(kpts):
            if self._use_cache:
                self.evecs = np.memmap(os.path.join(self.cache_path,
                                                    'evecs.dat'),
                                       mode='w+',
                                       shape=(nkpts, self.nbasis, self.nbasis),
                                       dtype=complex)
                self.H = np.memmap(os.path.join(self.cache_path, 'H.dat'),
                                   mode='w+',
                                   shape=(nkpts, self.nbasis, self.nbasis),
                                   dtype=complex)
                self.S = np.memmap(os.path.join(self.cache_path, 'H.dat'),
                                   mode='w+',
                                   shape=(nkpts, self.nbasis, self.nbasis),
                                   dtype=complex)
            self.H[ik] = Hk
            self.S[ik] = Sk
            self.evals[ik] = evalue
            self.evecs[ik] = evec
            if self._use_cache:
                del self.evecs
                del self.H
                del self.S
        return self.H, self.S, self.evals, self.evecs


    def _prepare_cache(self, path='./TB2J_results/cache'):
        if not os.path.exists(path):
            os.makedirs(path)
        else:
            os.remove(path)
        self.cache_path = path


    # def get_HSE_cached(self, kpt, convention=2):
    #     kpt = tuple(kpt)
    #     kname = f"{kpt[0]:9.5f}_{kpt[1]:9.5f}_{kpt[2]:9.5f}"
    #     if kname in self.cache_dict:
    #         path = self.cache_dict[kname]
    #         Hk = np.memmap(os.path.join(path, 'H.dat'),
    #                        dtype='complex',
    #                        mode='r',
    #                        shape=(self.nbasis, self.nbasis))
    #         Sk = np.memmap(os.path.join(path, 'H.dat'),
    #                        dtype='complex',
    #                        mode='r',
    #                        shape=(self.nbasis, self.nbasis))
    #         evalue = np.memmap(os.path.join(path, 'evalue.dat'),
    #                            dtype='float',
    #                            mode='r',
    #                            shape=(self.nbasis))
    #         evec = np.memmap(os.path.join(path, 'evec.dat'),
    #                          dtype='complex',
    #                          mode='r',
    #                          shape=(self.nbasis, self.nbasis))
    #     else:
    #         Hk, Sk, evalue, evec = self.get_HSE(kpt, convention=convention)
    #         path = os.path.join(self.cache_path, )
    #         self.cache_dict[tuple(kpt)] = path
    #         fpH = np.memmap(os.path.join(path, 'H.dat'))
    #     return Hk, Sk, evalue, evec

    def get_HSE(self, kpt, convention=2):
        Hk = self.Hk(kpt, convention=convention)
        Sk = self.Sk(kpt, convention=convention)
        evalue, evec = self.solve(kpt, convention=convention)
        return Hk, Sk, evalue, evec

    def get_fermi_level(self):
        return 0.0


#def test():
#    fdf = sisl.get_sile('/home/hexu/projects/learn_siesta/SMO_Wannier/siesta.fdf')
#    H = fdf.read_hamiltonian(order='nc',dim=2)
#    print(H._spin._is_polarized)
#    print(H.__dict__.keys())
#    print(H._geometry.__dict__.keys())
#    print(H._geometry.xyz)
#    print(H._geometry.sc)
#    H._geometry.sc.cell
#    orb=H._geometry._atoms[0].orbital[0]
#    print(orb.name())
#    s=SislWrapper(H)
#    s.view_info()
#
#import sisl
#fdf = sisl.get_sile('/home/hexu/projects/learn_siesta/SMO_Wannier/siesta.fdf')
#H = fdf.read_hamiltonian(order='nc',dim=2)
#if __name__=='__main__':
#    test()



"""
Units and constants.
"""
import numpy as np

ps = 1e-15
fs = 1e-12

bohr_mag = 9.27400995e-24
gyromagnetic_ratio = 1.76e11
mu_B = bohr_mag

#from scipy.constants import (Boltzmann, hbar, mu_0, epsilon_0,

Boltzmann, hbar, mu_0, epsilon_0, elementary_charge, pi = (
    1.38064852e-23, 1.0545718001391127e-34, 1.2566370614359173e-06,
    8.854187817620389e-12, 1.6021766208e-19, np.pi)

kb=Boltzmann

eV = elementary_charge * 1
meV = 0.001 * eV

# Pauli Matrices
sigma_x = np.array([[0, 1], [1, 0]])
sigma_y = np.array([[0, -1j], [1j, 0]])
sigma_z = np.array([[1, 0], [0, -1]])

# Creation/anihination operators
creation_op = np.array([[0, 0], [1, 0]])
annihilation_op = np.array([[0, 1], [0, 0]])

number_op = np.array([[0, 0], [0, 1]])



#!/usr/bin/env python3
import argparse
import os
import sys
from TB2J.Jdownfolder import JDownfolder_pickle


def main():
    parser = argparse.ArgumentParser(
        description=
        "TB2J_downfold: Downfold the exchange parameter with ligand spin as independent variable to that only with metal site spin as independent variables."
    )
    parser.add_argument('--inpath',
                        '-i',
                        type=str,
                        help='The path of the input TB2J results before the downfolding.',
                        default="TB2J_results")

    parser.add_argument(
        '--outpath',
        '-o',
        type=str,
        help='The path of the output TB2J results path after the downfolding',
        default="TB2J_results_downfolded",
    )

    parser.add_argument("--metals",
                        "-m",
                        type=str,
                        help='List of magnetic cation elements',
                        nargs='+',
                        default=[])

    parser.add_argument("--ligands",
                        "-l",
                        type=str,
                        help='List of ligand elements',
                        nargs='+',
                        default=[])

    args = parser.parse_args()

    if len(args.metals) == []:
        print("List of magnetic cation elements cannot be empty")

    if len(args.ligands) == []:
        print("List of ligand elements cannot be empty")

    JDownfolder_pickle(inpath=args.inpath,
                       metals=args.metals,
                       ligands=args.ligands,
                       outpath=args.outpath)


main()



#!/usr/bin/env python3
from TB2J.manager import gen_exchange_siesta
from TB2J.versioninfo import print_license
import sys
import argparse


def run_siesta2J():
    print_license()
    parser = argparse.ArgumentParser(
        description="siesta2J: Using magnetic force theorem to calculate exchange parameter J from siesta Hamiltonian"
    )
    parser.add_argument('--fdf_fname',
                        help="path of the input fdf file",
                        default='./',
                        type=str)
    parser.add_argument('--elements',
                        help="list of elements to be considered in Heisenberg model. For each element, a postfixes can be used to specify the orbitals(Only with Siesta backend), eg. Fe_3d, or Fe_3d_4s ",
                        default=None,
                        type=str,
                        nargs='*')
    parser.add_argument(
        '--rcut',
        help='range of R. The default is all the commesurate R to the kmesh',
        default=None,
        type=float)
    parser.add_argument('--efermi',
                        help='Fermi energy in eV',
                        default=None,
                        type=float)
    parser.add_argument(
        '--kmesh',
        help='kmesh in the format of kx ky kz. Monkhorst pack. If all the numbers are odd, it is Gamma cenetered. (strongly recommended)',
        type=int,
        nargs='*',
        default=[5, 5, 5])
    parser.add_argument('--emin',
                        help='energy minimum below efermi, default -14 eV',
                        type=float,
                        default=-14.0)
    parser.add_argument('--emax',
                        help='energy maximum above efermi, default 0.0 eV',
                        type=float,
                        default=0.05)
    parser.add_argument(
        '--use_cache',
        help="whether to use disk file for temporary storing wavefunctions and hamiltonian to reduce memory usage. Default: False",
        action='store_true',
        default=False)
    # parser.add_argument(
    #    '--height',
    #    help=
    #    'energy contour, a small number (often between 0.1 to 0.5, default 0.2)',
    #    type=float,
    #    default=0.1)
    parser.add_argument('--nz',
                        help='number of integration steps, default: 50',
                        default=50,
                        type=int)
    # parser.add_argument(
    #    '--nz2', help='number of steps 2, default: 200', default=200, type=int)
    # parser.add_argument(
    #    '--nz3', help='number of steps 3, default: 50', default=50, type=int)
    parser.add_argument(
        '--cutoff',
        help="The minimum of J amplitude to write, (in eV), default is 1e-5 eV",
        default=1e-5,
        type=float)

    parser.add_argument(
        '--exclude_orbs',
        help="the indices of wannier functions to be excluded from magnetic site. counting start from 0",
        default=[],
        type=int,
        nargs='+')

    parser.add_argument('--np',
                        help='number of cpu cores to use in parallel, default: 1',
                        default=1,
                        type=int)

    parser.add_argument(
        "--description",
        help="add description of the calculatiion to the xml file. Essential information, like the xc functional, U values, magnetic state should be given.",
        type=str,
        default="Calculated with TB2J.")

    parser.add_argument("--fname",
                        default='exchange.xml',
                        type=str,
                        help='exchange xml file name. default: exchange.xml')

    parser.add_argument(
        "--output_path",
        help="The path of the output directory, default is TB2J_results",
        type=str,
        default="TB2J_results")

    args = parser.parse_args()

    if args.elements is None:
        print("Please input the magnetic elements, e.g. --elements Fe Ni")
        sys.exit()

    include_orbs = {}
    for element in args.elements:
        if "_" in element:
            elem = element.split("_")[0]
            orb = element.split("_")[1:]
            include_orbs[elem] = orb
        else:
            include_orbs[element] = None

    gen_exchange_siesta(
        fdf_fname=args.fdf_fname,
        kmesh=args.kmesh,
        magnetic_elements=list(include_orbs.keys()),
        include_orbs=include_orbs,
        Rcut=args.rcut,
        emin=args.emin,
        emax=args.emax,
        nz=args.nz,
        # height=args.height,
        # nz1=args.nz1,
        # nz2=args.nz2,
        # nz3=args.nz3,
        description=args.description,
        output_path=args.output_path,
        use_cache=args.use_cache,
        np=args.np,
        exclude_orbs=args.exclude_orbs)


if __name__ == "__main__":
    run_siesta2J()



#!/usr/bin/env python3
import argparse
from TB2J.rotate_atoms import rotate_atom_xyz, rotate_xyz, check_ftype


def main():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument('fname',
                        help="name of the file containing the atomic structure",
                        type=str)
    parser.add_argument('--ftype',
                        help="type of the output files, e.g.  xyz. Please use the ",
                        default='xyz',
                        type=str)

    args=parser.parse_args()
    rotate_xyz(args.fname, ftype=args.ftype)


if __name__ == "__main__":
    main()



import os
from TB2J.myTB import MyTB, merge_tbmodels_spin
import numpy as np
from TB2J.exchange import ExchangeCL, ExchangeNCL
from TB2J.exchangeCL2 import ExchangeCL2
from TB2J.exchange_qspace import ExchangeCLQspace
from TB2J.utils import read_basis, auto_assign_basis_name
from ase.io import read
from TB2J.sisl_wrapper import SislWrapper
from TB2J.gpaw_wrapper import GPAWWrapper
from TB2J.wannier import parse_atoms

def gen_exchange(path,
                 colinear=True,
                 groupby='spin',
                 posfile=None,
                 prefix_up='wannier90.up',
                 prefix_dn='wannier90.dn',
                 prefix_SOC='wannier90',
                 min_hopping_norm=1e-4,
                 max_distance=None,
                 efermi=0,
                 magnetic_elements=[],
                 kmesh=[4, 4, 4],
                 emin=-12.0,
                 emax=0.0,
                 nz=100,
                 height=0.2,
                 nz1=50,
                 nz2=200,
                 nz3=50,
                 exclude_orbs=[],
                 Rcut=None,
                 ne=None,
                 use_cache=False,
                 np=1,
                 output_path='TB2J_results',
                 wannier_type="wannier90",
                 qspace=False,
                 description=''):
    try:
        fname=os.path.join(path, posfile)
        print(f"Reading atomic structure from file {fname}.")
        atoms = read(os.path.join(path, posfile))
    except Exception:

        print(f"Cannot read atomic structure from file {fname}. Trying to read from Wannier input.")
        if colinear:
            fname=os.path.join(path, f"{prefix_up}.win")
        else:
            fname=os.path.join(path, f"{prefix_SOC}.win")

        print(f"Reading atomic structure from file {fname}.")
        atoms= parse_atoms(fname)

    basis_fname = os.path.join(path, 'basis.txt')
    if colinear:
        if wannier_type.lower() == "wannier90":
            print("Reading Wannier90 hamiltonian: spin up.")
            tbmodel_up = MyTB.read_from_wannier_dir(path=path,
                                                    prefix=prefix_up,
                                                    atoms=atoms, 
                                                    nls=False)
            print("Reading Wannier90 hamiltonian: spin down.")
            tbmodel_dn = MyTB.read_from_wannier_dir(path=path,
                                                    prefix=prefix_dn,
                                                    atoms=atoms,
                                                    nls=False)
            if os.path.exists(basis_fname):
                basis = read_basis(basis_fname)
            else:
                basis, _ = auto_assign_basis_name(tbmodel_up.xred, atoms)
        elif wannier_type.lower() == "banddownfolder":
            print("Reading Banddownfolder hamiltonian: spin up.")
            tbmodel_up = MyTB.load_banddownfolder(path=path,
                                                  prefix=prefix_up,
                                                  atoms=atoms,                                                   nls=False)
            print("Reading Banddownfolder hamiltonian: spin down.")
            tbmodel_dn = MyTB.load_banddownfolder(path=path,
                                                  prefix=prefix_dn,
                                                  atoms=atoms, 
                                                  nls=False)

            basis, _ = auto_assign_basis_name(tbmodel_up.xred, atoms)
        else:
            raise ValueError(
                "wannier_type should be Wannier90 or banddownfolder.")

        print("Starting to calculate exchange.")
        description = f""" Input from collinear Wannier90 data.
 Tight binding data from {path}. 
 Prefix of wannier function files:{prefix_up} and {prefix_dn}.
Warning: Please check if the noise level of Wannier function Hamiltonian to make sure it is much smaller than the exchange values.
\n"""
        if not qspace:
            exchange = ExchangeCL2(tbmodels=(tbmodel_up, tbmodel_dn),
                                   atoms=atoms,
                                   basis=basis,
                                   efermi=efermi,
                                   magnetic_elements=magnetic_elements,
                                   kmesh=kmesh,
                                   emin=emin,
                                   emax=emax,
                                   nz=nz,
                                   height=height,
                                   nz1=nz1,
                                   nz2=nz2,
                                   nz3=nz3,
                                   exclude_orbs=exclude_orbs,
                                   Rcut=Rcut,
                                   ne=ne,
                                   np=np,
                                   use_cache=use_cache,
                                   description=description)
        else:
            exchange = ExchangeCLQspace(tbmodels=(tbmodel_up, tbmodel_dn),
                                        atoms=atoms,
                                        basis=basis,
                                        efermi=efermi,
                                        magnetic_elements=magnetic_elements,
                                        kmesh=kmesh,
                                        emin=emin,
                                        emax=emax,
                                        nz=nz,
                                        height=height,
                                        nz1=nz1,
                                        nz2=nz2,
                                        nz3=nz3,
                                        exclude_orbs=exclude_orbs,
                                        Rcut=Rcut,
                                        ne=ne,
                                        np=np,
                                        use_cache=use_cache,
                                        description=description)

        exchange.run(path=output_path)
        print(
            "All calculation finsihed. The results are in TB2J_results directory."
        )

    elif colinear and wannier_type.lower() == "banddownfolder":
        print("Reading Wannier90 hamiltonian: spin up.")
        tbmodel_up = MyTB.read_from_wannier_dir(path=path,
                                                prefix=prefix_up,
                                                atoms=atoms,
                                                groupby=None,
                                                nls=False)
        print("Reading Wannier90 hamiltonian: spin down.")
        tbmodel_dn = MyTB.read_from_wannier_dir(path=path,
                                                prefix=prefix_dn,
                                                atoms=atoms,
                                                groupby=None,
                                                nls=False)
        tbmodel = merge_tbmodels_spin(tbmodel_up, tbmodel_dn)
        basis, _ = auto_assign_basis_name(tbmodel.xred, atoms)
        description = f""" Input from collinear BandDownfolder data.
 Tight binding data from {path}. 
 Prefix of wannier function files:{prefix_up} and {prefix_dn}.
Warning: Please check if the noise level of Wannier function Hamiltonian to make sure it is much smaller than the exchange values.
\n"""
        print("Starting to calculate exchange.")
        exchange = ExchangeCL2(tbmodels=tbmodel,
                               atoms=atoms,
                               basis=basis,
                               efermi=efermi,
                               magnetic_elements=magnetic_elements,
                               kmesh=kmesh,
                               emin=emin,
                               emax=emax,
                               nz=nz,
                               height=height,
                               nz1=nz1,
                               nz2=nz2,
                               nz3=nz3,
                               exclude_orbs=exclude_orbs,
                               Rcut=Rcut,
                               ne=ne,
                               np=np,
                               use_cache=use_cache,
                               description=description)
        exchange.run(path=output_path)
        print(
            "All calculation finsihed. The results are in TB2J_results directory."
        )
    else:
        print("Reading Wannier90 hamiltonian: non-colinear spin.")
        groupby = groupby.lower().strip()
        if groupby not in ['spin', 'orbital']:
            raise ValueError("groupby can only be spin or orbital.")
        tbmodel = MyTB.read_from_wannier_dir(path=path,
                                             prefix=prefix_SOC,
                                             atoms=atoms,
                                             groupby=groupby,
                                             nls=True)
        if os.path.exists(basis_fname):
            print("The use of basis file is deprecated. It will be ignored.")
            #basis = read_basis(basis_fname)
        else:
            basis, _ = auto_assign_basis_name(tbmodel.xred, atoms)
        description = f""" Input from non-collinear Wannier90 data.
 Tight binding data from {path}. 
 Prefix of wannier function files:{prefix_SOC}.
Warning: Please check if the noise level of Wannier function Hamiltonian to make sure it is much smaller than the exchange values.
 The DMI component parallel to the spin orientation, the Jani which has the component of that orientation should be disregarded
 e.g. if the spins are along z, the xz, yz, zz, zx, zy components and the z component of DMI.
 If you need these component, try to do three calculations with spin along x, y, z,  or use structure with z rotated to x, y and z. And then use TB2J_merge.py to get the full set of parameters.

\n"""
        print("Starting to calculate exchange.")
        exchange = ExchangeNCL(tbmodels=tbmodel,
                               atoms=atoms,
                               basis=basis,
                               efermi=efermi,
                               magnetic_elements=magnetic_elements,
                               kmesh=kmesh,
                               emin=emin,
                               emax=emax,
                               nz=nz,
                               height=height,
                               nz1=nz1,
                               nz2=nz2,
                               nz3=nz3,
                               exclude_orbs=exclude_orbs,
                               Rcut=Rcut,
                               ne=ne,
                               np=np,
                               use_cache=use_cache,
                               description=description)
        print("\n")
        exchange.run(path=output_path)
        print(
            f"All calculation finsihed. The results are in {output_path} directory."
        )


def gen_exchange_siesta(fdf_fname,
                        magnetic_elements=[],
                        include_orbs=None,
                        kmesh=[5, 5, 5],
                        emin=-12.0,
                        emax=0.0,
                        nz=100,
                        exclude_orbs=[],
                        Rcut=None,
                        ne=None,
                        np=1,
                        use_cache=False,
                        output_path='TB2J_results',
                        description=''):

    try:
        import sisl
    except:
        raise ImportError(
            "sisl cannot be imported. Please install sisl first.")

    from packaging import version
    if version.parse(sisl.__version__) <= version.parse("0.10.0"):
        raise ImportError(
            f"sisl version is {sisl.__version__}, but should be larger than 0.10.0."
        )
    fdf = sisl.get_sile(fdf_fname)
    H = fdf.read_hamiltonian()
    if H.spin.is_colinear:
        print("Reading Siesta hamiltonian: colinear spin.")
        tbmodel_up = SislWrapper(H, spin=0)
        tbmodel_dn = SislWrapper(H, spin=1)
        basis = dict(zip(tbmodel_up.orbs, list(range(tbmodel_up.norb))))
        print("Starting to calculate exchange.")
        description = f""" Input from collinear Siesta data.
 working directory: {os.getcwd()}
 fdf_fname: {fdf_fname}.
\n"""
        exchange = ExchangeCL2(
            tbmodels=(tbmodel_up, tbmodel_dn),
            atoms=tbmodel_up.atoms,
            basis=basis,
            efermi=0.0,
            magnetic_elements=magnetic_elements,
            include_orbs=include_orbs,
            kmesh=kmesh,
            emin=emin,
            emax=emax,
            nz=nz,
            #height=height,
            #nz1=nz1,
            #nz2=nz2,
            #nz3=nz3,
            exclude_orbs=exclude_orbs,
            Rcut=Rcut,
            ne=ne,
            np=np,
            use_cache=use_cache,
            description=description)
        exchange.run(path=output_path)
        print("\n")
        print(
            f"All calculation finsihed. The results are in {output_path} directory."
        )

    elif H.spin.is_colinear:
        print(
            "Reading Siesta hamiltonian: colinear spin. Treat as non-colinear")
        tbmodel = SislWrapper(H, spin='merge')
        basis = dict(zip(tbmodel.orbs, list(range(tbmodel.nbasis))))
        print("Starting to calculate exchange.")
        description = f""" Input from collinear Siesta data.
 working directory: {os.getcwd()}
 fdf_fname: {fdf_fname}.
\n"""
        exchange = ExchangeNCL(tbmodels=tbmodel,
                               atoms=tbmodel.atoms,
                               basis=basis,
                               efermi=0.0,
                               magnetic_elements=magnetic_elements,
                               include_orbs=include_orbs,
                               kmesh=kmesh,
                               emin=emin,
                               emax=emax,
                               nz=nz,
                               exclude_orbs=exclude_orbs,
                               Rcut=Rcut,
                               ne=ne,
                               np=np,
                               use_cache=use_cache,
                               description=description)
        exchange.run(path=output_path)
        print("\n")
        print(
            f"All calculation finsihed. The results are in {output_path} directory."
        )

    elif H.spin.is_spinorbit or H.spin.is_noncolinear:

        print("Reading Siesta hamiltonian: non-colinear spin.")
        tbmodel = SislWrapper(H, spin=None)
        basis = dict(zip(tbmodel.orbs, list(range(tbmodel.nbasis))))
        print("Starting to calculate exchange.")
        description = f""" Input from non-collinear Siesta data.
 working directory: {os.getcwd()}
 fdf_fname: {fdf_fname}.
Warning: The DMI component parallel to the spin orientation, the Jani which has the component of that orientation should be disregarded
 e.g. if the spins are along z, the xz, yz, zz, zx, zy components and the z component of DMI.
 If you need these component, try to do three calculations with spin along x, y, z,  or use structure with z rotated to x, y and z. And then use TB2J_merge.py to get the full set of parameters.
\n"""
        exchange = ExchangeNCL(tbmodels=tbmodel,
                               atoms=tbmodel.atoms,
                               basis=basis,
                               efermi=0.0,
                               magnetic_elements=magnetic_elements,
                               include_orbs=include_orbs,
                               kmesh=kmesh,
                               emin=emin,
                               emax=emax,
                               nz=nz,
                               exclude_orbs=exclude_orbs,
                               Rcut=Rcut,
                               ne=ne,
                               np=np,
                               use_cache=use_cache,
                               description=description)
        exchange.run(path=output_path)
        print("\n")
        print(
            f"All calculation finsihed. The results are in {output_path} directory."
        )


def gen_exchange_gpaw(gpw_fname,
                      magnetic_elements=[],
                      kmesh=[3, 3, 3],
                      emin=-12.0,
                      emax=0.0,
                      nz=50,
                      exclude_orbs=[],
                      Rcut=None,
                      use_cache=False,
                      output_path='TB2J_results',
                      description=''):
    print("Reading from GPAW data and calculate electronic structure.")
    model = GPAWWrapper(gpw_fname=gpw_fname)
    efermi = model.calc.get_fermi_level()
    print(f"Fermi Energy: {efermi}")
    poses = np.vstack([model.positions, model.positions])
    basis, _ = auto_assign_basis_name(poses, model.atoms)
    if model.calc.get_spin_polarized():
        print("Starting to calculate exchange.")
        exchange = ExchangeNCL(tbmodels=model,
                               atoms=model.atoms,
                               efermi=efermi,
                               basis=basis,
                               magnetic_elements=magnetic_elements,
                               kmesh=kmesh,
                               emin=emin,
                               emax=emax,
                               nz=nz,
                               exclude_orbs=exclude_orbs,
                               Rcut=Rcut,
                               use_cache=use_cache,
                               description=description)
        exchange.run(path=output_path)
        print("\n")
        print(
            f"All calculation finsihed. The results are in {output_path} directory."
        )



import numpy as np
import xml.etree.cElementTree as ET
from xml.dom import minidom
from ase.data import atomic_masses
from ase.units import eV, Hartree, Bohr, Ry, J
import os
from collections import Iterable, namedtuple
from itertools import groupby
from TB2J import __version__
from TB2J.utils import symbol_number
from numpy import array_str
from TB2J.spinham.spin_api import SpinModel
from datetime import datetime
import pickle


def write_info_section(cls, myfile):
    myfile.write('=' * 90 + '\n')
    myfile.write("Information: \n")
    now = datetime.now()
    #myfile.write("Exchange parameters generated by TB2J %s\n" % (__version__))
    #myfile.write("Generation time: %s\n" % (now.strftime("%y/%m/%d %H:%M:%S")))
    myfile.write(cls.description)


def write_atom_section(cls, myfile):
    """
    write the atom section
    including the cell and the atomic positions.
    """
    myfile.write('=' * 90 + '\n')
    myfile.write("Cell (Angstrom):\n")
    cell = cls.atoms.get_cell()
    for c in cell:
        myfile.write("{:6.3f}  {:6.3f}  {:6.3f}\n".format(c[0], c[1], c[2]))

    # write atom information
    myfile.write('\n')
    myfile.write('=' * 90 + '\n')
    myfile.write('Atoms:  \n')
    myfile.write(
        '(Note: charge and magmoms only count the wannier functions.)\n')
    if cls.colinear:
        myfile.write("{:^12s} {:^9s} {:^9s} {:^9s} {:^9s} {:^9s}\n".format(
            'Atom number', 'x', 'y', 'z', 'w_charge', 'w_magmom'))
    else:
        myfile.write(
            "{:^12s} {:^9s} {:^9s} {:^9s} {:^9s} {:^9s} {:^9s} {:^9s}\n".
            format('Atom_number', 'x', 'y', 'z', 'w_charge', 'M(x)', 'M(y)',
                   'M(z)'))

    symnum = symbol_number(cls.atoms)
    sns = list(symnum.keys())
    poses = cls.atoms.get_positions()

    tchg, tmx, tmy, tmz = 0, 0, 0, 0
    for i, s in enumerate(symnum):
        if cls.colinear:
            mag = cls.magmoms[i]
            chg = cls.charges[i]
            myfile.write(
                "{:<12s} {:9.4f} {:9.4f} {:9.4f} {:9.4f} {:9.4f}\n".format(
                    s, poses[i, 0], poses[i, 1], poses[i, 2], chg, mag))
        else:
            chg = cls.charges[i]
            mx, my, mz = cls.spinat[i, :]
            myfile.write(
                "{:<12s} {:9.4f} {:9.4f} {:9.4f} {:9.4f} {:9.4f} {:9.4f} {:9.4f}\n"
                .format(s, poses[i, 0], poses[i, 1], poses[i, 2], chg, mx, my,
                        mz))
            tchg += chg
            tmx += mx
            tmy += my
            tmz += mz
    if cls.colinear:
        myfile.write("{:<12s} {:9s} {:9s} {:9s} {:9.4f} {:9.4f}\n".format(
            'Total', '', '', '', np.sum(cls.charges), np.sum(cls.magmoms)))
    else:
        myfile.write(
            "{:<12s} {:9s} {:9s} {:9s} {:9.4f} {:9.4f} {:9.4f} {:9.4f}\n".
            format('Total', '', '', '', tchg, tmx, tmy, tmz))

    myfile.write('\n')


def write_orbital_section(cls, myfile):
    if cls.exchange_Jdict_orb is not None:
        myfile.write('=' * 90 + '\n')
        myfile.write('Orbitals used in decomposition: \n')
        myfile.write("The name of the orbitals for the decomposition: \n")
        symnum = symbol_number(cls.atoms)
        sns = list(symnum.keys())
        for iatom in cls.orbital_names:
            if cls.index_spin[iatom] != -1:
                myfile.write(f"{sns[iatom]} : {cls.orbital_names[iatom]}\n")
        myfile.write("\n")


def write_exchange_section(cls,
                           myfile,
                           order='distance',
                           write_experimental: bool = True):
    symnum = symbol_number(cls.atoms)
    sns = list(symnum.keys())
    poses = cls.atoms.get_positions()

    myfile.write('=' * 90 + '\n')
    myfile.write('Exchange: \n')

    # l = [x for x in cls.exchange_Jlist if abs(x.J) > cutoff]
    keys = cls.exchange_Jdict
    if order == 'amplitude':
        l = sorted(keys,
                   key=lambda x: abs(cls.exchange_Jdict[x]),
                   reverse=True)
    elif order == 'distance':
        l = sorted(keys, key=lambda x: cls.distance_dict[x][1])
    else:
        l = keys

    myfile.write("{:6s} {:5s} {:15s} {:7s} {:24s} {:11s} \n".format(
        "    i", "    j", "         R", "  J_iso(meV)", "         vector",
        "distance(A)"))

    for ll in l:
        myfile.write('-' * 88 + '\n')
        R, i, j = ll
        J = cls.exchange_Jdict[ll]
        vec, distance = cls.distance_dict[ll]
        myfile.write(
            "   {:5s} {:5s} ({:3d}, {:3d}, {:3d}) {:7.4f}   ({:6.3f}, {:6.3f}, {:6.3f}) {:6.3f} \n"
            .format(sns[cls.ind_atoms[i]], sns[cls.ind_atoms[j]], R[0], R[1],
                    R[2], J * 1e3, vec[0], vec[1], vec[2], distance))

        Jiso = cls.exchange_Jdict[ll] * 1e3
        myfile.write(f'J_iso: {Jiso:7.4f} \n')

        if cls.has_biquadratic and write_experimental:
            Jprime, B = cls.biquadratic_Jdict[ll]
            myfile.write(
                f"[Testing!] Jprime: {Jprime*1e3:.3f},  B: {B*1e3:.3f}\n")

        if cls.dJdx is not None:
            dJdx = cls.dJdx[ll]
            myfile.write(f"dJ/dx: {dJdx*1e3:.3f}\n")

        if cls.dmi_ddict is not None:
            DMI = cls.dmi_ddict[ll] * 1e3
            myfile.write(
                '[Testing!] DMI: ({:7.4f} {:7.4f} {:7.4f})\n'.format(
                    DMI[0], DMI[1], DMI[2]))

        if write_experimental:
            try:
                #DMI2 = cls.debug_dict['DMI2'][ll] * 1e3
                #myfile.write('[Debug!] DMI2: ({:7.4f} {:7.4f} {:7.4f})\n'.format(
                #    DMI2[0], DMI2[1], DMI2[2]))
                pass
            except:
                pass

        if cls.Jani_dict is not None and write_experimental:
            J = cls.Jani_dict[ll] * 1e3
            myfile.write(
                f"[Testing!]J_ani:\n{array_str(J, precision=3, suppress_small=True)}\n"
            )

        if cls.NJT_ddict is not None:
            DMI = cls.NJT_ddict[ll] * 1e3
            myfile.write(
                '[Experimental!] DMI_NJt: ({:7.4f} {:7.4f} {:7.4f})\n'.format(
                    DMI[0], DMI[1], DMI[2]))

        if cls.NJT_Jdict is not None:
            J = cls.NJT_Jdict[ll] * 1e3
            myfile.write(
                '[Testing!] Jani_NJt: ({:7.4f} {:7.4f} {:7.4f})\n'.format(
                    J[0], J[1], J[2]))

        if cls.exchange_Jdict_orb:
            myfile.write("Orbital contributions:\n {} \n".format(
                np.array_str(cls.exchange_Jdict_orb[ll] * 1e3,
                             precision=3,
                             suppress_small=True)))


def write_Jq_info(cls, kpts, evals, evecs, myfile, special_kpoints={}):
    """
    kpts: the list of kpoints
    evals: eigen values
    evecs: eigen vectors
    fname: the name of the output file
    spk: special kpoints
    """
    symnum = symbol_number(cls.atoms)
    sns = list(symnum.keys())
    imin = np.argmin(evals[:, 0])
    #emin = np.min(evals[:, 0])
    nspin = evals.shape[1] // 3
    evec_min = evecs[imin, :, 0].reshape(nspin, 3)

    # write information to file
    if myfile is not None:
        myfile.write("\nThe energy minimum is at the q-point:")
        myfile.write("%s\n" % kpts[np.argmin(evals[:, 0])])
        myfile.write("The eigenstate at the energy minimum is:\n")
        for i, ev in enumerate(evec_min):
            v = ev.real / np.linalg.norm(ev)
            myfile.write(" > %s: (%.3f, %.3f, %.3f)\n" %
                         (sns[cls.ind_atoms[i]], v[0], v[1], v[2]))
    print("\nThe energy minimum is at:")
    print("%s\n" % kpts[np.argmin(evals[:, 0])])
    print("The eigenstate at the energy minimum is:\n")
    for i, ev in enumerate(evec_min):
        v = ev.real / np.linalg.norm(ev)
        print("spin %s: (%.3f, %.3f, %.3f)" %
              (sns[cls.ind_atoms[i]], v[0], v[1], v[2]))


def write_txt(cls,
              path='TB2J_results',
              fname='exchange.out',
              order='distance',
              write_orb=False,
              write_experimental=True,
              cutoff=1e-4):
    if not os.path.exists(path):
        os.makedirs(path)
    fname = os.path.join(path, fname)
    with open(fname, 'w') as myfile:
        write_info_section(cls, myfile)
        write_atom_section(cls, myfile)
        write_orbital_section(cls, myfile)
        write_exchange_section(cls,
                               myfile,
                               write_experimental=write_experimental)



from collections import defaultdict, OrderedDict
import os
import numpy as np
from TB2J.green import TBGreen
from TB2J.pauli import (pauli_block_all, pauli_block_sigma_norm, pauli_mat)
from TB2J.utils import symbol_number, read_basis, kmesh_to_R
from TB2J.myTB import MyTB
from ase.io import read
from TB2J.utils import auto_assign_basis_name
from TB2J.io_exchange import SpinIO
from tqdm import tqdm
from p_tqdm import p_map
from functools import lru_cache
from TB2J.contour import Contour
from TB2J.utils import simpson_nonuniform, trapezoidal_nonuniform
from TB2J.orbmap import map_orbs_matrix
from pathos.multiprocessing import ProcessPool


class Exchange():
    def __init__(
            self,
            tbmodels,
            atoms,
            efermi,
            basis=None,
            magnetic_elements=[],
            include_orbs={},
            kmesh=[4, 4, 4],
            emin=-15,  # integration lower bound, relative to fermi energy
            # integration upper bound. Should be 0 (fermi energy). But DFT codes define Fermi energy in various ways.
            emax=0.05,
            nz=100,
            # the delta in the (i delta) in green's function to prevent divergence
            height=0.5,
            nz1=150,  # grid from emin to emin+(i delta)
            nz2=300,  # grid from emin +(i delta) to emax+(i delta)
            nz3=150,  # grid from emax + (i delta) to emax
            exclude_orbs=[],  #
            ne=None,  # number of electrons in Wannier function.
            Rcut=None,  # Rcut.
            use_cache=False,
            np=1,
            description=''):

        self.atoms = atoms
        self.efermi = efermi
        self.emin = emin
        self.emax = emax
        self.nz = nz
        self.height = height
        self.nz1 = nz1
        self.nz2 = nz2
        self.nz3 = nz3
        if nz is None:
            self.nz = nz1 + nz2 + nz3
        self._prepare_kmesh(kmesh)
        self.Rcut = Rcut
        self.basis = basis
        self.magnetic_elements = magnetic_elements
        self.include_orbs = include_orbs

        self.exclude_orbs = exclude_orbs
        self.ne = ne
        self._use_cache = use_cache
        self.np = np

        self.set_tbmodels(tbmodels)
        self._adjust_emin()
        self._prepare_elist(method='legendre')
        self._prepare_Rlist()
        self._prepare_basis()
        self._prepare_orb_dict()
        self._prepare_distance()

        # whether to calculate J and DMI with NJt method.
        self.calc_NJt = True
        # self._prepare_NijR()
        self.Ddict_NJT = None
        self.Jdict_NJT = None
        self._is_collinear = True
        self.has_elistc = False
        self.description = description
        self._clean_tbmodels()

    def _adjust_emin(self):
        self.emin = self.G.find_energy_ingap(rbound=self.efermi -
                                             5.0) - self.efermi

    def set_tbmodels(self, tbmodels):
        pass

    def _clean_tbmodels(self):
        del self.tbmodel
        del self.G.tbmodel

    def _prepare_kmesh(self, kmesh):
        for k in kmesh:
            self.kmesh = list(map(lambda x: x // 2 * 2 + 1, kmesh))

    def _prepare_elist(self, method='legendre'):
        """
        prepare list of energy for integration.
        The path has three segments:
         emin --1-> emin + 1j*height --2-> emax+1j*height --3-> emax
        """
        self.contour = Contour(self.emin, self.emax)
        if method.lower() == 'rectangle':
            self.contour.build_path_rectangle(height=self.height,
                                              nz1=self.nz1,
                                              nz2=self.nz2,
                                              nz3=self.nz3)
        elif method.lower() == 'semicircle':
            self.contour.build_path_semicircle(npoints=self.nz, endpoint=True)
        elif method.lower() == 'legendre':
            self.contour.build_path_legendre(npoints=self.nz, endpoint=True)
        else:
            raise ValueError(f"The path cannot be of type {method}.")

    def _prepare_Rlist(self):
        """
        prepare R list for J(i, j, R)
        [-Rx, Rx] * [-Ry, Ry] * [-Rz, Rz]
        """
        self.Rlist = kmesh_to_R(self.kmesh)

    def _prepare_basis(self):
        if self.basis is None:
            pass

    def _prepare_orb_dict(self):
        """
        generate self.ind_mag_atoms and self.orb_dict
        """
        # adict: dictionary of {'Fe': ['dxy', 'dyz', ...], ...}
        adict = OrderedDict()
        # orb_dict: {ind_atom:[ind_orb,1,2], ...}
        self.orb_dict = {}
        # labels: {0:{dxy, ...}}
        self.labels = {}
        # magnetic atoms index
        self.ind_mag_atoms = []

        sdict = symbol_number(self.atoms)

        for i, base in enumerate(self.basis):
            if i not in self.exclude_orbs:
                # e.g. Fe2, dxy, _, _
                if isinstance(base, str):
                    atom_sym, orb_sym = base.split('|')[:2]
                else:
                    atom_sym, orb_sym = base[:2]

                if atom_sym in adict:
                    adict[atom_sym].append(orb_sym)
                else:
                    adict[atom_sym] = [orb_sym]
                iatom = sdict[atom_sym]
                if iatom not in self.orb_dict:
                    self.orb_dict[iatom] = [i]
                    self.labels[iatom] = [orb_sym]
                else:
                    self.orb_dict[iatom] += [i]
                    self.labels[iatom] += [orb_sym]

        #self.orb_slice = []

        #for iatom in range(len(self.atoms)):
        #    if iatom in self.orb_dict:
        #        self.orb_slice.append(
        #            slice(
        #                self.orb_dict[iatom][0],
        #                self.orb_dict[iatom][-1] + 1,
        #            ))
        #    else:
        #        self.orb_slice.append(slice(0, 0))

        #self.orb_slice = np.array(self.orb_slice)

        # index of magnetic atoms
        for i, sym in enumerate(self.atoms.get_chemical_symbols()):
            if sym in self.magnetic_elements:
                self.ind_mag_atoms.append(i)

        self._spin_dict = {}
        self._atom_dict = {}
        for ispin, iatom in enumerate(self.ind_mag_atoms):
            self._spin_dict[iatom] = ispin
            self._atom_dict[ispin] = iatom

        self._prepare_orb_mmat()

    def _prepare_orb_mmat(self):
        self.mmats = {}
        self.orbital_names = {}
        if self.backend_name == "SIESTA":
            syms = self.atoms.get_chemical_symbols()
            for iatom, orbs in self.labels.items():
                if (self.include_orbs is not None) and syms[iatom] in self.include_orbs:
                    mmat, reduced_orbs = map_orbs_matrix(
                        orbs,
                         spinor=not (self._is_collinear),
                         include_only=self.include_orbs[syms[iatom]])
                else:
                    mmat, reduced_orbs = map_orbs_matrix(
                        orbs,
                         spinor=not (self._is_collinear),
                         include_only=None)

                self.mmats[iatom] = mmat
                self.orbital_names[iatom] = reduced_orbs
        else:
            self.orbital_names = self.labels

    def ispin(self, iatom):
        return self._spin_dict[iatom]

    def iatom(self, ispin):
        return self._atom_dict[ispin]

    def _prepare_distance(self):
        self.distance_dict = {}
        self.short_Rlist = []
        self.R_ijatom_dict = defaultdict(lambda: [])
        ind_matoms = self.ind_mag_atoms
        for ispin, iatom in enumerate(ind_matoms):
            for jspin, jatom in enumerate(ind_matoms):
                for R in self.Rlist:
                    pos_i = self.atoms.get_positions()[iatom]
                    pos_jR = self.atoms.get_positions()[jatom] + np.dot(
                        R, self.atoms.get_cell())
                    vec = pos_jR - pos_i
                    distance = np.sqrt(np.sum(vec**2))
                    if self.Rcut is None or distance < self.Rcut:
                        self.distance_dict[(tuple(R), ispin,
                                            jspin)] = (vec, distance)
                        self.R_ijatom_dict[tuple(R)].append((iatom, jatom))
        self.short_Rlist = list(self.R_ijatom_dict.keys())

    def iorb(self, iatom):
        """
        given an index of atom, return all the orbital indices of it.
        :param iatom:  index of atom.
        """
        return np.array(self.orb_dict[iatom], dtype=int)

    def calculate_all(self):
        raise NotImplementedError(
            "calculate all is not implemented in a abstract class. ")

    def write_output(self):
        raise NotImplementedError(
            "write_output is not implemented in a abstract class. ")


class ExchangeNCL(Exchange):
    """
    Non-collinear exchange
    """
    def set_tbmodels(self, tbmodels):
        """
        tbmodels should be in spinor form.
        The basis should be orb1_up, orb2_up,...orbn_up, orb1_dn, orb2_dn....
        """
        self.tbmodel = tbmodels
        self.backend_name = self.tbmodel.name
        # TODO: check if tbmodels are really a tbmodel with SOC.
        self.G = TBGreen(self.tbmodel,
                         self.kmesh,
                         self.efermi,
                         use_cache=self._use_cache,
                         nproc=self.np)
        self.norb = self.G.norb
        self.nbasis = self.G.nbasis
        self.rho = np.zeros((self.nbasis, self.nbasis), dtype=complex)
        self.A_ijR_list = defaultdict(lambda: [])
        self.A_ijR = defaultdict(lambda: np.zeros((4, 4), dtype=complex))
        self.HR0 = self.G.H0
        self._is_collinear = False
        self.Pdict = {}

    def _prepare_NijR(self):
        self.N = {}
        for R in self.Rlist:
            self.N[R] = np.zeros((self.nbasis, self.nbasis), dtype=complex)

    def _prepare_Patom(self):
        for iatom in self.ind_mag_atoms:
            self.Pdict[iatom] = pauli_block_sigma_norm(self.get_H_atom(iatom))

    def get_H_atom(self, iatom):
        orbs = self.iorb(iatom)
        #return self.HR0[self.orb_slice[iatom], self.orb_slice[iatom]]
        return self.HR0[np.ix_(orbs, orbs)]

    def get_P_iatom(self, iatom):
        """ Calculate the norm of the Hamiltonian vector.
        For each atom, the local hamiltonian of each orbital H(2*2 matrix)
        can be written as H0* I + H1* sigma1 + H2*sigma2 + H3 *sigma3
        where sigma is a Pauli matrix. return the norm of (H1, H2, H3) vector
        :param iatom: index of atom
        :returns: a matrix of norms P. P[i, j] is for orbital i and j.
        :rtype: complex matrix of shape norb_i * norb_i.
        """
        if self.Pdict == {}:
            self._prepare_Patom()
        return self.Pdict[iatom]

    def GR_atom(self, GR, iatom, jatom):
        """Given a green's function matrix, return the [iatom, jatom] component.

        :param GR:  Green's function matrix
        :param iatom:  index of atom i
        :param jatom:  index of atom j
        :returns:  G_ij
        :rtype:  complex matrix.

        """
        orbi = self.iorb(iatom)
        orbj = self.iorb(jatom)
        return GR[np.ix_(orbi, orbj)]
        #return GR[self.orb_slice[iatom], self.orb_slice[jatom]]

    def get_A_ijR(self, G, R, iatom, jatom):
        """ calculate A from G for a energy slice (de).
        It take the
        .. math::
           A^{uv} = p T^u p T^v dE / pi

        where u, v are I, x, y, z (index 0, 1,2,3). p(i) = self.get_P_iatom(iatom)
        T^u(ijR)  (u=0,1,2,3) = pauli_block_all(G)

        :param G: Green's function for all R, i, j.
        :param iatom: i
        :param jatom: j
        :param de:  energy step. used for integeration
        :returns: a matrix of A_ij(u, v), where u, v =(0)0, x(1), y(2), z(3)
        :rtype:  4*4 matrix
        """
        GR = G[R]
        Gij = self.GR_atom(GR, iatom, jatom)
        Gij_Ixyz = pauli_block_all(Gij)

        # G(j, i, -R)
        Rm = tuple(-x for x in R)
        GRm = G[Rm]
        Gji = self.GR_atom(GRm, jatom, iatom)
        Gji_Ixyz = pauli_block_all(Gji)

        tmp = np.zeros((4, 4), dtype=complex)
        for a in range(4):
            pGp = self.get_P_iatom(iatom) @ Gij_Ixyz[a] @ self.get_P_iatom(
                jatom)
            for b in range(4):
                AijRab = np.matmul(pGp, Gji_Ixyz[b])
                tmp[a, b] = np.trace(AijRab)
        return tmp / np.pi

    def get_all_A(self, G):
        """
        Calculate all A matrix elements
        Loop over all magnetic atoms.
        :param G: Green's function.
        :param de: energy step.
        """
        A_ijR_list = {}
        for iR, R in enumerate(self.R_ijatom_dict):
            for (iatom, jatom) in self.R_ijatom_dict[R]:
                A = self.get_A_ijR(G, R, iatom, jatom)
                A_ijR_list[(R, iatom, jatom)] = A
        return A_ijR_list

    def A_to_Jtensor(self):
        """
        Calculate J tensors from A.
        If we assume the exchange can be written as a bilinear tensor form,
        J_{isotropic} = Tr Im (A^{00} - A^{xx} - A^{yy} - A^{zz})
        J_{anisotropic}_uv = Tr Im (2A)
        DMI =  Tr Re (A^{0z} - A^{z0} )
        """
        self.Jani = {}
        self.DMI = {}

        self.Jprime = {}
        self.B = {}
        self.exchange_Jdict = {}
        self.debug_dict = {'DMI2': {}}
        for key, val in self.A_ijR.items():
            # key:(R, iatom, jatom)
            R, iatom, jatom = key

            Rm = tuple(-x for x in R)
            valm = self.A_ijR[(Rm, jatom, iatom)]

            ispin = self.ispin(iatom)
            jspin = self.ispin(jatom)
            keyspin = (R, ispin, jspin)

            is_nonself = not (R == (0, 0, 0) and iatom == jatom)
            Jiso = np.zeros((3, 3), dtype=float)
            Ja = np.zeros((3, 3), dtype=float)
            Dtmp = np.zeros(3, dtype=float)
            Dtmp2 = np.zeros(3, dtype=float)
            # Heisenberg like J.
            for i in range(3):
                Jiso[i, i] += np.imag(val[0, 0] - val[1, 1] - val[2, 2] -
                                      val[3, 3])

            if is_nonself:
                self.exchange_Jdict[keyspin] = Jiso[0, 0]

            # off-diagonal anisotropic exchange
            for i in range(3):
                for j in range(3):
                    Ja[i, j] = np.imag(val[i + 1, j + 1] + valm[i + 1, j + 1])
            if is_nonself:
                self.Jani[keyspin] = Ja

            # DMI
            for i in range(3):
                Dtmp[i] = np.real(val[0, i + 1] - val[i + 1, 0])

            # Dx = Jyz-Jzy
            # Dy = Jzx-Jxz
            # Dz = Jxy-Jyx
            Dtmp2[0] = np.imag(val[2, 3] - val[3, 2])
            Dtmp2[1] = np.imag(val[3, 1] - val[1, 3])
            Dtmp2[2] = np.imag(val[1, 2] - val[2, 1])
            if is_nonself:
                self.DMI[keyspin] = Dtmp
                self.debug_dict['DMI2'][keyspin] = Dtmp2

            # isotropic exchange into bilinear and biqudratic parts:
            # Jprime SiSj and B (SiSj)^2
            if is_nonself:
                Si = self.spinat[iatom]
                Sj = self.spinat[jatom]
                Jprime = np.imag(val[0, 0] - val[3, 3]) - 2 * np.sign(
                    np.dot(Si, Sj)) * np.imag(val[3, 3])
                # Jprime = np.imag(val[0, 0] - 3*val[3, 3])
                B = np.imag(val[3, 3])
                self.B[keyspin] = Jprime, B

    def get_N_e(self, GR, de):
        """
        calcualte density matrix for all R,i, j
        """
        self.N = defaultdict(lambda: 0.0)
        for R, G in GR.items():
            self.N[R] += -1.0 / np.pi * np.imag(G * de)

    def get_rho_e(self, rhoR):
        """ add component to density matrix from a green's function
        :param GR: Green's funciton in real space.
        """
        return -1.0 / np.pi * rhoR[0, 0, 0]

    def get_total_charges(self):
        return np.sum(np.imag(np.diag(self.rho)))

    def get_rho_atom(self):
        """
        calculate charge and spin for each atom.
        """
        rho = {}
        self.charges = np.zeros(len(self.atoms), dtype=float)
        self.spinat = np.zeros((len(self.atoms), 3), dtype=float)
        for iatom in self.orb_dict:
            iorb = self.iorb(iatom)
            tmp = self.rho[np.ix_(iorb, iorb)]
            # *2 because there is a 1/2 in the paui_block_all function
            rho[iatom] = np.array(
                [np.trace(x) * 2 for x in pauli_block_all(tmp)])
            self.charges[iatom] = np.imag(rho[iatom][0])
            self.spinat[iatom, :] = np.imag(rho[iatom][1:])
        self.rho_dict = rho
        return self.rho_dict

    def calculate_DMI_NJT(self):
        """
        calculate exchange and DMI with the
        D(i,j) =
        """
        Ddict_NJT = {}
        Jdict_NJT = {}
        for R in self.short_Rlist:
            N = self.N[tuple(-np.array(R))]  # density matrix
            t = self.tbmodel.get_hamR(R)  # hopping parameter
            for iatom in self.ind_mag_atoms:
                orbi = self.iorb(iatom)
                ni = len(orbi)
                for jatom in self.ind_mag_atoms:
                    orbj = self.iorb(jatom)
                    nj = len(orbj)
                    Nji = N[np.ix_(orbj, orbi)]
                    tij = t[np.ix_(orbi, orbj)]
                    D = np.zeros(3, dtype=float)
                    J = np.zeros(3, dtype=float)
                    for dim in range(3):
                        # S_i = pauli_mat(ni, dim +
                        #                1)  #*self.rho[np.ix_(orbi, orbi)]
                        # S_j = pauli_mat(nj, dim +
                        #                1)  #*self.rho[np.ix_(orbj, orbj)]
                        # TODO: Note that rho is complex, not the imaginary part
                        S_i = pauli_mat(ni, dim + 1) * self.rho[np.ix_(
                            orbi, orbi)]
                        S_j = pauli_mat(nj, dim + 1) * self.rho[np.ix_(
                            orbj, orbj)]

                        # [S, t]+  = Si tij + tij Sj, where
                        # Si and Sj are the spin operator
                        # Here we do not have L operator, so J-> S
                        Jt = np.matmul(S_i, tij) + np.matmul(tij, S_j)

                        Jtminus = np.matmul(S_i, tij) - np.matmul(tij, S_j)
                        # D = -1/2 Tr Nji [J, tij]
                        # Trace over spin and orb
                        D[dim] = -0.5 * np.imag(np.trace(np.matmul(Nji, Jt)))
                        J[dim] = -0.5 * np.imag(
                            np.trace(np.matmul(Nji, Jtminus)))
                    ispin = self.ispin(iatom)
                    jspin = self.ispin(jatom)
                    Ddict_NJT[(R, ispin, jspin)] = D
                    Jdict_NJT[(R, ispin, jspin)] = J
        self.Jdict_NJT = Jdict_NJT
        self.Ddict_NJT = Ddict_NJT
        return Ddict_NJT

    def integrate(self, rhoRs, AijRs, method='simpson'):
        """
        AijRs: a list of AijR, 
        wherer AijR: array of ((nR, n, n, 4,4), dtype=complex)
        """
        if method == "trapezoidal":
            integrate = trapezoidal_nonuniform
        elif method == 'simpson':
            integrate = simpson_nonuniform

        self.rho = integrate(self.contour.path, rhoRs)
        for iR, R in enumerate(self.R_ijatom_dict):
            for (iatom, jatom) in self.R_ijatom_dict[R]:
                f = AijRs[(R, iatom, jatom)]
                self.A_ijR[(R, iatom, jatom)] = integrate(self.contour.path, f)

    def get_AijR_rhoR(self, e):
        GR, rhoR = self.G.get_GR(self.short_Rlist, energy=e, get_rho=True)
        AijR = self.get_all_A(GR)
        return AijR, self.get_rho_e(rhoR)

    def save_AijR(self, AijRs, fname):
        result = dict(path=self.contour.path, AijRs=AijRs)
        with open(fname, 'wb') as myfile:
            pickle.dump(result, myfile)

    def calculate_all(self):
        """
        The top level.
        """
        print("Green's function Calculation started.")

        rhoRs = []
        GRs = []
        AijRs = {}

        npole=len(self.contour.path)
        if self.np > 1:
            #executor = ProcessPool(nodes=self.np)
            #results = executor.map(self.get_AijR_rhoR, self.contour.path)
            results = p_map(self.get_AijR_rhoR, self.contour.path, num_cpus=self.np)
        else:
            results = map(self.get_AijR_rhoR, tqdm(self.contour.path, total=npole))

        for i, result in enumerate(results):
            for iR, R in enumerate(self.R_ijatom_dict):
                for (iatom, jatom) in self.R_ijatom_dict[R]:
                    if (R, iatom, jatom) in AijRs:
                        AijRs[(R, iatom, jatom)].append(result[0][R, iatom,
                                                                  jatom])
                    else:
                        AijRs[(R, iatom, jatom)] = []
                        AijRs[(R, iatom, jatom)].append(result[0][R, iatom,
                                                                  jatom])
            rhoRs.append(result[1])
        if self.np > 1:
            #executor.close()
            #executor.join()
            #executor.clear()
            pass

        # self.save_AijRs(AijRs)
        self.integrate(rhoRs, AijRs)

        self.get_rho_atom()
        self.A_to_Jtensor()

    def _prepare_index_spin(self):
        # index_spin: index in spin hamiltonian of atom. starts from 1. -1 means not considered.
        ind_matoms = []
        self.index_spin = []
        ispin = 0
        for i, sym in enumerate(self.atoms.get_chemical_symbols()):
            if sym in self.magnetic_elements:
                ind_matoms.append(i)
                self.index_spin.append(ispin)
                ispin += 1
            else:
                self.index_spin.append(-1)

    def write_output(self, path='TB2J_results'):
        self._prepare_index_spin()
        output = SpinIO(
            atoms=self.atoms,
            charges=self.charges,
            spinat=self.spinat,
            index_spin=self.index_spin,
            colinear=False,
            orbital_names=self.orbital_names,
            distance_dict=self.distance_dict,
            exchange_Jdict=self.exchange_Jdict,
            dmi_ddict=self.DMI,
            NJT_Jdict=self.Jdict_NJT,
            NJT_ddict=self.Ddict_NJT,
            Jani_dict=self.Jani,
            biquadratic_Jdict=self.B,
            debug_dict=self.debug_dict,
            description=self.description,
        )
        output.write_all(path=path)

    def finalize(self):
        self.G.clean_cache()

    def run(self, path='TB2J_results'):
        self.calculate_all()
        self.write_output(path=path)
        self.finalize()


class ExchangeCL(ExchangeNCL):
    def set_tbmodels(self, tbmodel):
        """
        only difference is a colinear tag.
        """
        super().set_tbmodels(tbmodel)
        self._is_collinear = True

    def write_output(self, path='TB2J_results'):
        self._prepare_index_spin()
        output = SpinIO(
            atoms=self.atoms,
            charges=self.charges,
            spinat=self.spinat,
            index_spin=self.index_spin,
            orbital_names=self.orbital_names,
            colinear=True,
            distance_dict=self.distance_dict,
            exchange_Jdict=self.exchange_Jdict,
            dmi_ddict=None,
            NJT_Jdict=None,
            NJT_ddict=None,
            biquadratic_Jdict=self.B,
            description=self.description,
        )
        output.write_all(path=path)



import numpy as np
import scipy.linalg as sl
from collections import defaultdict
from ase.dft.kpoints import monkhorst_pack
from shutil import rmtree
import os
import tempfile
from pathos.multiprocessing import ProcessPool
import sys


def eigen_to_G(evals, evecs, efermi, energy):
    """ calculate green's function from eigenvalue/eigenvector for energy(e-ef): G(e-ef).
    :param evals:  eigen values
    :param evecs:  eigen vectors
    :param efermi: fermi energy
    :param energy: energy
    :returns: Green's function G,
    :rtype:  Matrix with same shape of the Hamiltonian (and eigenvector)
    """
    return np.einsum("ij, j-> ij", evecs, 1.0 /
                     (-evals + (energy + efermi))) @ evecs.conj().T
    #return np.einsum("ij, j, jk -> ik", evecs, 1.0 / (-evals + (energy + efermi)), evecs.conj().T)
    #return evecs.dot(np.diag(1.0 / (-evals + (energy + efermi)))).dot(
    #    evecs.conj().T)


MAX_EXP_ARGUMENT = np.log(sys.float_info.max)


def fermi(e, mu, width=0.01):
    """
    the fermi function.
     .. math::
        f=\\frac{1}{\exp((e-\mu)/width)+1}

    :param e,mu,width: e,\mu,width
    """

    x = (e - mu) / width
    return np.where(x < MAX_EXP_ARGUMENT, 1 / (1.0 + np.exp(x)), 0.0)


def find_energy_ingap(evals, rbound, gap=1.0):
    """
    find a energy inside a gap below rbound (right bound), 
    return the energy gap top - 0.5.
    """
    m0 = np.sort(evals.flatten())
    m = m0[m0 < rbound]
    ind = np.where(np.diff(m) > gap)[0]
    if len(ind) == 0:
        return m0[0] - 0.5
    else:
        return m[ind[-1] + 1] - 0.5


class TBGreen():
    def __init__(
        self,
        tbmodel,
        kmesh,  # [ikpt, 3]
        efermi,  # efermi
        k_sym=False,
        use_cache=False,
        cache_path=None,
        nproc=1):
        """
        :param tbmodel: A tight binding model
        :param kmesh: size of monkhorst pack. e.g [6,6,6]
        :param efermi: fermi energy.
        """
        self.tbmodel = tbmodel
        self.is_orthogonal = tbmodel.is_orthogonal
        self.R2kfactor = tbmodel.R2kfactor
        self.k2Rfactor = -tbmodel.R2kfactor
        self.efermi = efermi
        self._use_cache = use_cache
        self.cache_path = cache_path
        if use_cache:
            self._prepare_cache()
        if kmesh is not None:
            self.kpts = monkhorst_pack(size=kmesh)
        else:
            self.kpts = tbmodel.get_kpts()
        self.nkpts = len(self.kpts)
        self.kweights = [1.0 / self.nkpts] * self.nkpts
        self.norb = tbmodel.norb
        self.nbasis = tbmodel.nbasis
        self.k_sym = k_sym
        self.nproc = nproc
        self._prepare_eigen()

    def _reduce_eigens(self, evals, evecs, emin, emax):
        ts = np.logical_and(evals >= emin, evals < emax)
        ts = np.any(ts, axis=0)
        ts = np.where(ts)[0]
        istart, iend = ts[0], ts[-1] + 1
        return evals[:, istart:iend], evecs[:, :, istart:iend]

    def find_energy_ingap(self, rbound, gap=1.0):
        return find_energy_ingap(self.evals, rbound, gap)

    def _prepare_cache(self):
        if self.cache_path is None:
            if 'TMPDIR' in os.environ:
                rpath = os.environ['TMPDIR']
            else:
                rpath = '/dev/shm/TB2J_cache'
        else:
            rpath = self.cache_path
        if not os.path.exists(rpath):
            os.makedirs(rpath)
        self.cache_path = tempfile.mkdtemp(prefix='TB2J', dir=rpath)
        print(
            f"Writting wavefunctions and Hamiltonian in cache {self.cache_path}"
        )

    def clean_cache(self):
        if (self.cache_path is not None) and os.path.exists(self.cache_path):
            rmtree(self.cache_path)

    def _prepare_eigen(self):
        """
        calculate eigen values and vectors for all kpts and save.
        Note that the convention 2 is used here, where the 
        phase factor is e^(ik.R), not e^(ik.(R+rj-ri))
        """
        nkpts = len(self.kpts)
        self.evals = np.zeros((nkpts, self.nbasis), dtype=float)
        self.nkpts = nkpts
        self.H0 = np.zeros((self.nbasis, self.nbasis), dtype=complex)
        self.evecs = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
        H = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
        if not self.is_orthogonal:
            self.S = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
        else:
            self.S = None
        if self.nproc == 1:
            results = map(self.tbmodel.HSE_k, self.kpts)
        else:
            executor = ProcessPool(nodes=self.nproc)
            results = executor.map(self.tbmodel.HSE_k, self.kpts,
                                   [2] * len(self.kpts))
            executor.close()
            executor.join()
            executor.clear()

        for ik, result in enumerate(results):
            if self.is_orthogonal:
                H[ik], _, self.evals[ik], self.evecs[ik] = result
            else:
                H[ik], self.S[ik], self.evals[ik], self.evecs[ik] = result
            self.H0 += H[ik] / self.nkpts

        self.evals, self.evecs = self._reduce_eigens(self.evals,
                                                     self.evecs,
                                                     emin=self.efermi - 10.0,
                                                     emax=self.efermi + 10.1)
        if self._use_cache:
            evecs = self.evecs
            self.evecs_shape = self.evecs.shape
            self.evecs = np.memmap(os.path.join(self.cache_path, 'evecs.dat'),
                                   mode='w+',
                                   shape=self.evecs.shape,
                                   dtype=complex)
            self.evecs[:, :, :] = evecs[:, :, :]
            if self.is_orthogonal:
                self.S = None
            else:
                S = self.S
                self.S = np.memmap(os.path.join(self.cache_path, 'S.dat'),
                                   mode='w+',
                                   shape=(nkpts, self.nbasis, self.nbasis),
                                   dtype=complex)
                self.S[:] = S[:]
            del self.evecs
            if not self.is_orthogonal:
                del self.S

    def get_evecs(self, ik):
        if self._use_cache:
            return np.memmap(os.path.join(self.cache_path, 'evecs.dat'),
                             mode='r',
                             shape=self.evecs_shape,
                             dtype=complex)[ik]
        else:
            return self.evecs[ik]

    def get_evalue(self, ik):
        return self.evals[ik]

    def get_Hk(self, ik):
        if self._use_cache:
            return np.memmap(os.path.join(self.cache_path, 'H.dat'),
                             mode='r',
                             shape=(self.nkpts, self.nbasis, self.nbasis),
                             dtype=complex)[ik]
        else:
            return self.evecs[ik]

    def get_Sk(self, ik):
        if self.is_orthogonal:
            return None
        elif self._use_cache:
            return np.memmap(os.path.join(self.cache_path, 'S.dat'),
                             mode='r',
                             shape=(self.nkpts, self.nbasis, self.nbasis),
                             dtype=complex)[ik]
        else:
            return self.S[ik]

    def get_density_matrix(self):
        rho = np.zeros((self.nbasis, self.nbasis), dtype=complex)
        for ik, _ in enumerate(self.kpts):
            rho += (self.get_evecs(ik) * fermi(self.evals[ik], self.efermi)
                    ) @ self.get_evecs(ik).T.conj() * self.kweights[ik]
        return rho

    def get_density(self):
        return np.real(np.diag(self.get_density_matrix()))

    def get_Gk(self, ik, energy):
        """ Green's function G(k) for one energy
        G(\epsilon)= (\epsilon I- H)^{-1}
        :param ik: indices for kpoint
        :returns: Gk
        :rtype:  a matrix of indices (nbasis, nbasis)
        """
        Gk = eigen_to_G(evals=self.get_evalue(ik),
                        evecs=self.get_evecs(ik),
                        efermi=self.efermi,
                        energy=energy)

        # A slower version. For test.
        #Gk = np.linalg.inv((energy+self.efermi)*self.S[ik,:,:] - self.H[ik,:,:])
        return Gk

    def get_GR(self, Rpts, energy, get_rho=False):
        """ calculate real space Green's function for one energy, all R points.
        G(R, epsilon) = G(k, epsilon) exp(-2\pi i R.dot. k)
        :param Rpts: R points
        :param energy:
        :returns:  real space green's function for one energy for a list of R.
        :rtype:  dictionary, the keys are tuple of R, values are matrices of nbasis*nbasis
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0j)
        rhoR = defaultdict(lambda: 0.0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            if get_rho:
                if self.is_orthogonal:
                    rhok = Gk
                else:
                    rhok = self.get_Sk(ik) @ Gk
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                tmp = Gk * (phase * self.kweights[ik])
                GR[R] += tmp
                # change this if need full rho
                if get_rho and R == (0, 0, 0):
                    rhoR[R] += rhok * (phase * self.kweights[ik])
        if get_rho:
            return GR, rhoR
        else:
            return GR

    def get_GR_and_dGRdx1(self, Rpts, energy, dHdx):
        """
        calculate G(R) and dG(R)/dx.
        dG(R)/dx = \sum_k G(k) (dH(R)/dx) G(k).
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0 + 0.0j)
        dGRdx = defaultdict(lambda: 0.0 + 0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            Gkw = Gk * self.kweights[ik]
            #Gmk = self.get_Gk(self.i_minus_k(kpt), energy)
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                GR[R] += Gkw * (phase * self.kweights[ik])

                dHRdx = dHdx.get_hamR(R)
                dGRdx[R] += Gkw @ dHRdx @ Gk
                #dGRdx[R] += Gk.dot(dHRdx).dot(Gkp)
        return GR, dGRdx

    def get_GR_and_dGRdx(self, Rpts, energy, dHdx):
        """
        calculate G(R) and dG(R)/dx.
        dG(k)/dx =  G(k) (dH(k)/dx) G(k).
        dG(R)/dx = \sum_k dG(k)/dx * e^{-ikR}
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0 + 0.0j)
        dGRdx = defaultdict(lambda: 0.0 + 0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            #Gmk = self.get_Gk(self.i_minus_k(kpt), energy)
            Gkp = Gk * self.kweights[ik]
            dHk = dHdx.gen_ham(tuple(kpt))
            dG = Gk @ dHk @ Gkp
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                GR[R] += Gkp * (phase * self.kweights[ik])
                dGRdx[R] += dG * (phase * self.kweights[ik])
        return GR, dGRdx

    def get_GR_and_dGRdx_and_dGRdx2(self, Rpts, energy, dHdx, dHdx2):
        """
        calculate G(R) and dG(R)/dx.
        dG(k)/dx =  G(k) (dH(k)/dx) G(k).
        dG(R)/dx = \sum_k dG(k)/dx * e^{-ikR}
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0 + 0.0j)
        dGRdx = defaultdict(lambda: 0.0 + 0j)
        dGRdx2 = defaultdict(lambda: 0.0 + 0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            #Gmk = self.get_Gk(self.i_minus_k(kpt), energy)
            Gkp = Gk * self.kweights[ik]
            dHk = dHdx.gen_ham(tuple(kpt))
            dHk2 = dHdx2.gen_ham(tuple(kpt))
            dG = Gk @ dHk @ Gkp
            dG2 = Gk @ dHk2 @ Gkp
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                GR[R] += Gkp * (phase * self.kweights[ik])
                dGRdx[R] += dG * (phase * self.kweights[ik])
                dGRdx2[R] += dG2 * (phase * self.kweights[ik])
        return GR, dGRdx, dGRdx2



import numpy as np
from .hamiltonian import SpinHamiltonian, read_spin_ham_from_file
#from minimulti.spin.mover import SpinMover


class SpinModel():
    def __init__(self, fname=None, sc_matrix=None):
        if fname is not None:
            self.read_from_file(fname)
        else:
            self._ham = SpinHamiltonian()

        if sc_matrix is not None:
            self.make_supercell(sc_matrix)

        #self.mover = SpinMover(self._ham)

    @property
    def ham(self):
        return self._ham

    def add_term(self, term, name=None):
        self._ham.add_Hamiltonian_term(term, name=name)

    @ham.setter
    def ham(self, ham):
        self._ham = ham

    #@property
    #def S(self):
    #    return self.mover.s

    @property
    def nspin(self):
        return self._ham.nspin

    def read_from_file(self, fname):
        self._ham = read_spin_ham_from_file(fname)

    def set(self, **kwargs):
        self.params.set(**kwargs)
        # self.mover.set(
        #     time_step=self.params.time_step,
        #     temperature=self.params.temperature,
        #     total_time=self.params.total_time,
        # )

    def set_ham(self, **kwargs):
        self._ham.set(**kwargs)

    def make_supercell(self, sc_matrix=None, supercell_maker=None):
        self._ham = self._ham.make_supercell(
            sc_matrix=sc_matrix, supercell_maker=supercell_maker)
        #self.mover = SpinMover(self._ham)
        return self

    # def run_one_step(self):
    #     self.mover.run_one_step()

    # def run_time(self):
    #     self.mover.run()

    def plot_magnon_band(
            self,
            kvectors=np.array([[0, 0, 0], [0.5, 0, 0], [0.5, 0.5, 0],
                               [0, 0, 0], [.5, .5, .5]]),
            knames=['$\Gamma$', 'X', 'M', '$\Gamma$', 'R'],
            supercell_matrix=None,
            npoints=100,
            color='red',
            ax=None,
            kpath_fname=None,
            Jq=False,
    ):
        self._ham.plot_magnon_band(
            kvectors=kvectors,
            knames=knames,
            supercell_matrix=supercell_matrix,
            npoints=npoints,
            color=color,
            ax=ax,
            Jq=Jq,
            kpath_fname=kpath_fname)



import unittest
import numpy as np
from TB2J.myTB2 import MyTB
from TB2J.pauli import *  #pauli_block
from TB2J.green_SOC import TBGreen
from TB2J.utils import auto_assign_basis_name
from TB2J.exchange import ExchangeNCL, gen_exchange


class AssignTest(unittest.TestCase):
    def test_assign(self):
        pass


class PauliTest(unittest.TestCase):
    def test_pauli(self):
        M = np.array([[1 - 3j, 9 + 2j], [3 - 9j, -4]])
        #print(np.sum(M * s1T) * s1)
        M0, Mx, My, Mz = pauli_decomp(M)
        #print(M0, Mx, My, Mz)
        #print(sum([M0 * s0, Mx * s1, My * s2, Mz * s3]))

        MI = (pauli_block_I(M, 1))
        Mx = (pauli_block_x(M, 1))
        My = (pauli_block_y(M, 1))
        Mz = (pauli_block_z(M, 1))

        n = 8
        hn = n // 2
        M = np.random.random([n, n]) + 0.0j
        MI = (pauli_block_I(M, hn))
        Mx = (pauli_block_x(M, hn))
        My = (pauli_block_y(M, hn))
        Mz = (pauli_block_z(M, hn))


class TBTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path = '/media/hexu/Backup/materials/SOC/RuCl3_FM_SOC/'
        #m = MyTB.read_from_wannier_dir(
        #    path=self.path, prefix='wannier90', poscar='POSCAR', nls=True)
        #m.save('RuCl3.nc')
        cls.tbmodel = MyTB.load_MyTB('RuCl3.nc')
        cls.positions = cls.tbmodel.xred
        cls.atoms = cls.tbmodel.atoms

    def test_read_wannier(self):
        m = MyTB.read_from_wannier_dir(
            path=self.path, prefix='wannier90', poscar='POSCAR', nls=True)
        m.save('RuCl3.nc')
        pass

    def test_saveload_wannier(self):
        # test onsite energy
        Eonsite = self.tbmodel.onsite_energies
        self.assertEqual(first=Eonsite[0], second=-5.028565)
        self.assertAlmostEqual(Eonsite[-1], -6.9492, delta=0.0001)

        # test R=0
        ER0 = self.tbmodel.ham_R0
        self.assertAlmostEqual(ER0[0, 0], -5.028565, delta=0.0001)
        self.assertAlmostEqual(ER0[-1, -1], -6.9492, delta=0.0001)

        # test positive R
        ER100 = self.tbmodel.get_hamR(R=(1, 0, 0))
        self.assertAlmostEqual(ER100[0, 0], -0.000740, delta=0.0001)
        self.assertAlmostEqual(ER100[1, 0], -0.003922, delta=0.0001)

        # test negative R
        ERm100 = self.tbmodel.get_hamR(R=(-1, 0, 0))
        self.assertAlmostEqual(ERm100[0, 0], -0.000740, delta=0.0001)
        self.assertAlmostEqual(ERm100[1, 0], -0.002838, delta=0.0001)

    def test_assign_basis(self):
        basis_dict, shifted_pos = auto_assign_basis_name(
            positions=self.tbmodel.xred, atoms=self.atoms)
        # Note that it starts from 1.
        self.assertEqual(basis_dict['Ru1|orb_1'], 1)
        self.assertEqual(basis_dict['Ru1|orb_6'], 29)

    def test_Green(self):
        G = TBGreen(
            tbmodel=self.tbmodel,
            kmesh=[4, 4, 4],
            efermi=3.0,
        )
        G00 = G.get_Gk(ik=0, energy=0.0)
        self.assertAlmostEqual(G00[0, 0], 1.30602e-1, delta=1e-4)
        GR0 = G.get_GR(Rpts=[(0, 0, 0)], energy=0.0)
        self.assertAlmostEqual(GR0[(0, 0, 0)][0, 0], 0.130001329, delta=1e-4)


class ExchangeTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        path = '/media/hexu/Backup/materials/SOC/RuCl3_FM_SOC/'
        cls.exchange = gen_exchange(
            kmesh=[6, 6, 3],
            Rmesh=[1,1,1],
            nz1=50,
            nz3=50,
            nz2=300,
            path=path,
            efermi=-3.25,
            magnetic_elements=['Ru'])

    def test_orb_dict(self):
        print(self.exchange.orb_dict)

    def test_runall(self):
        self.exchange.calculate_all()
        print(self.exchange.Jtensor)
        print(self.exchange.DMI)

    def test_write_txt(self):
        self.exchange.write_output()


class IOTest(unittest.TestCase):
    def test_xml(self):
        pass

    def test_txt(self):
        pass


unittest.main()
